# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import eventlet

from neutron.agent.common import async_process


class OFEvent(object):

    def __init__(self, event_type, flow):
        self.event_type = event_type
        self.flow = flow


class OFMonitor(async_process.AsyncProcess):
    """Wrapper over 'ovs-ofctl monitor'.

    This is an interactive OpenFlow monitor. By default, when the object is
    instantiated, the monitor process is started. To retrieve the pending
    events, the property "of_events" can be retrieved.

    NOTE(ralonsoh): 'ovs-ofctl monitor' command is sending existing flows to
    stdout pipe (startup first messages) and next incoming messages to stderr
    pipe. That's why this     class joins both outputs in one single queue
    (self._queue).
    """

    EVENT_RE = re.compile(
        r"event=(?P<action>ADDED|DELETED|MODIFIED) (?P<flow>.*)")

    def __init__(self, bridge_name, namespace=None, respawn_interval=None,
                 start=True):
        cmd = ['ovs-ofctl', 'monitor', bridge_name, 'watch:', '--monitor']
        super(OFMonitor, self).__init__(cmd, run_as_root=True,
                                        respawn_interval=respawn_interval,
                                        namespace=namespace)
        if start:
            self.start()

        self._queue = eventlet.queue.Queue()
        eventlet.spawn(self._read_and_enqueue, self.iter_stdout)
        eventlet.spawn(self._read_and_enqueue, self.iter_stderr)

    def _read_and_enqueue(self, iter):
        for event_line in iter(block=True):
            event = self._parse_event_line(event_line)
            if event:
                self._queue.put(event)

    @property
    def of_events(self):
        events = []
        while not self._queue.empty():
            events.append(self._queue.get())
        return events

    def _parse_event_line(self, event_line):
        match = self.EVENT_RE.match(event_line)
        if match is None:
            return
        return OFEvent(match.group('action'), match.group('flow'))

    def start(self, **kwargs):
        if not self._is_running:
            super(OFMonitor, self).start(block=True)

    def stop(self, **kwargs):
        if self._is_running:
            super(OFMonitor, self).stop(block=True)



# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
from neutron_lib.api.definitions import portbindings as portbindings_api
from neutron_lib.api.definitions import qos as qos_api
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib.services.qos import constants as qos_constants
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.api import extensions
from neutron.common.ovn import constants as ovn_const
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.core_extensions import qos as core_qos
from neutron.objects import network as network_obj
from neutron.objects import ports as port_obj
from neutron.objects.qos import policy as policy_obj
from neutron.objects.qos import rule as rule_obj
from neutron.objects import router as router_obj
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb.extensions \
    import qos as qos_extension
from neutron.tests.unit.plugins.ml2 import test_plugin


QOS_RULE_BW_1 = {'max_kbps': 200, 'max_burst_kbps': 100}
QOS_RULE_BW_2 = {'max_kbps': 300}
QOS_RULE_DSCP_1 = {'dscp_mark': 16}
QOS_RULE_DSCP_2 = {'dscp_mark': 20}
QOS_RULE_MINBW_1 = {'min_kbps': 500}


class _Context(object):

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return


class TestOVNClientQosExtension(test_plugin.Ml2PluginV2TestCase):

    CORE_PLUGIN_CLASS = 'neutron.plugins.ml2.plugin.Ml2Plugin'
    _extension_drivers = [qos_api.ALIAS]
    l3_plugin = ('neutron.tests.unit.extensions.test_qos_fip.'
                 'TestFloatingIPQoSL3NatServicePlugin')

    def setUp(self):
        ovn_conf.register_opts()
        cfg.CONF.set_override('extension_drivers', self._extension_drivers,
                              group='ml2')
        cfg.CONF.set_override('enable_distributed_floating_ip', 'False',
                              group='ovn')
        extensions.register_custom_supported_check(qos_api.ALIAS, lambda: True,
                                                   plugin_agnostic=True)
        super(TestOVNClientQosExtension, self).setUp()
        self.setup_coreplugin(self.CORE_PLUGIN_CLASS, load_plugins=True)
        self._mock_qos_loaded = mock.patch.object(
            core_qos.QosCoreResourceExtension, 'plugin_loaded')
        self.mock_qos_loaded = self._mock_qos_loaded.start()
        self.txn = _Context()
        mock_driver = mock.Mock()
        mock_driver._nb_idl.transaction.return_value = self.txn
        self.qos_driver = qos_extension.OVNClientQosExtension(
            driver=mock_driver)
        self._mock_rules = mock.patch.object(self.qos_driver,
                                             '_update_port_qos_rules')
        self.mock_rules = self._mock_rules.start()
        self.addCleanup(self._mock_rules.stop)
        self.ctx = context.get_admin_context()
        self.project_id = uuidutils.generate_uuid()
        self._initialize_objs()

    def _get_random_db_fields(self, obj_cls=None):
        obj_cls = obj_cls or self._test_class
        return obj_cls.modify_fields_to_db(
            self.get_random_object_fields(obj_cls))

    def _create_one_port(self, mac_address_int, network_id):
        mac_address = netaddr.EUI(mac_address_int)
        port = port_obj.Port(
            self.ctx, project_id=self.project_id,
            network_id=network_id, device_owner='',
            admin_state_up=True, status='DOWN', device_id='2',
            mac_address=mac_address)
        port.create()
        return port

    def _create_one_router(self):
        self.router_gw_port = self._create_one_port(2000, self.fips_network.id)
        self.router = router_obj.Router(self.ctx, id=uuidutils.generate_uuid(),
                                        gw_port_id=self.router_gw_port.id)
        self.router.create()

    def _initialize_objs(self):
        self.qos_policies = []
        self.ports = []
        self.networks = []
        self.fips = []
        self.fips_network = network_obj.Network(
            self.ctx, id=uuidutils.generate_uuid(), project_id=self.project_id)
        self.fips_network.create()
        self._create_one_router()
        self.fips_ports = []
        fip_cidr = netaddr.IPNetwork('10.10.0.0/24')

        for net_idx in range(2):
            qos_policy = policy_obj.QosPolicy(
                self.ctx, id=uuidutils.generate_uuid(),
                project_id=self.project_id)
            qos_policy.create()
            self.qos_policies.append(qos_policy)

            # Any QoS policy should have at least one rule, in order to have
            # the port dictionary extended with the QoS policy information; see
            # QoSPlugin._extend_port_resource_request
            qos_rule = rule_obj.QosDscpMarkingRule(
                self.ctx, dscp_mark=20, id=uuidutils.generate_uuid(),
                qos_policy_id=qos_policy.id)
            qos_rule.create()

            self.fips_ports.append(self._create_one_port(1000 + net_idx,
                                                         self.fips_network.id))
            fip_ip = str(netaddr.IPAddress(fip_cidr.ip + net_idx + 1))
            fip = router_obj.FloatingIP(
                self.ctx, id=uuidutils.generate_uuid(),
                project_id=self.project_id, floating_ip_address=fip_ip,
                floating_network_id=self.fips_network.id,
                floating_port_id=self.fips_ports[-1].id)
            fip.create()
            self.fips.append(fip)

            network = network_obj.Network(
                self.ctx, id=uuidutils.generate_uuid(),
                project_id=self.project_id)
            network.create()
            self.networks.append(network)

            for port_idx in range(3):
                self.ports.append(
                    self._create_one_port(net_idx * 16 + port_idx, network.id))

    @mock.patch.object(qos_extension.LOG, 'warning')
    @mock.patch.object(rule_obj, 'get_rules')
    def test__qos_rules(self, mock_get_rules, mock_warning):
        rules = [
            rule_obj.QosBandwidthLimitRule(
                direction=constants.EGRESS_DIRECTION, **QOS_RULE_BW_1),
            rule_obj.QosBandwidthLimitRule(
                direction=constants.INGRESS_DIRECTION, **QOS_RULE_BW_2),
            rule_obj.QosDscpMarkingRule(**QOS_RULE_DSCP_1),
            rule_obj.QosMinimumBandwidthRule(**QOS_RULE_MINBW_1)]
        mock_get_rules.return_value = rules
        expected = {
            constants.EGRESS_DIRECTION: {
                qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_1,
                qos_constants.RULE_TYPE_DSCP_MARKING: QOS_RULE_DSCP_1},
            constants.INGRESS_DIRECTION: {
                qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_2}
        }
        self.assertEqual(expected, self.qos_driver._qos_rules(mock.ANY,
                                                              'policy_id1'))
        mock_warning.assert_not_called()

    @mock.patch.object(rule_obj, 'get_rules')
    def test__qos_rules_no_rules(self, mock_get_rules):
        mock_get_rules.return_value = []
        expected = {constants.EGRESS_DIRECTION: {},
                    constants.INGRESS_DIRECTION: {}}
        self.assertEqual(expected,
                         self.qos_driver._qos_rules(mock.ANY, mock.ANY))

    def _test__ovn_qos_rule_ingress(self, fip_id=None, ip_address=None):
        if fip_id:
            external_ids = {ovn_const.OVN_FIP_EXT_ID_KEY: fip_id}
        else:
            external_ids = {ovn_const.OVN_PORT_EXT_ID_KEY: 'port_id'}
        direction = constants.INGRESS_DIRECTION
        rule = {qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_1}
        match = self.qos_driver._ovn_qos_rule_match(
            direction, 'port_id', ip_address, 'resident_port')
        expected = {'burst': 100, 'rate': 200, 'direction': 'to-lport',
                    'match': match,
                    'priority': qos_extension.OVN_QOS_DEFAULT_RULE_PRIORITY,
                    'switch': 'neutron-network_id',
                    'external_ids': external_ids}
        result = self.qos_driver._ovn_qos_rule(
            direction, rule, 'port_id', 'network_id', fip_id=fip_id,
            ip_address=ip_address, resident_port='resident_port')
        self.assertEqual(expected, result)

    def test__ovn_qos_rule_ingress(self):
        self._test__ovn_qos_rule_ingress()

    def test__ovn_qos_rule_ingress_fip(self):
        self._test__ovn_qos_rule_ingress(fip_id='fipid', ip_address='1.2.3.4')

    def _test__ovn_qos_rule_egress(self, fip_id=None, ip_address=None):
        if fip_id:
            external_ids = {ovn_const.OVN_FIP_EXT_ID_KEY: fip_id}
        else:
            external_ids = {ovn_const.OVN_PORT_EXT_ID_KEY: 'port_id'}
        direction = constants.EGRESS_DIRECTION
        rule = {qos_constants.RULE_TYPE_DSCP_MARKING: QOS_RULE_DSCP_1}
        match = self.qos_driver._ovn_qos_rule_match(
            direction, 'port_id', ip_address, 'resident_port')
        expected = {'direction': 'from-lport', 'match': match,
                    'dscp': 16, 'switch': 'neutron-network_id',
                    'priority': qos_extension.OVN_QOS_DEFAULT_RULE_PRIORITY,
                    'external_ids': external_ids}
        result = self.qos_driver._ovn_qos_rule(
            direction, rule, 'port_id', 'network_id', fip_id=fip_id,
            ip_address=ip_address, resident_port='resident_port')
        self.assertEqual(expected, result)

        rule = {qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_2,
                qos_constants.RULE_TYPE_DSCP_MARKING: QOS_RULE_DSCP_2}
        expected = {'direction': 'from-lport', 'match': match,
                    'rate': 300, 'dscp': 20, 'switch': 'neutron-network_id',
                    'priority': qos_extension.OVN_QOS_DEFAULT_RULE_PRIORITY,
                    'external_ids': external_ids}
        result = self.qos_driver._ovn_qos_rule(
            direction, rule, 'port_id', 'network_id', fip_id=fip_id,
            ip_address=ip_address, resident_port='resident_port')
        self.assertEqual(expected, result)

    def test__ovn_qos_rule_egress(self):
        self._test__ovn_qos_rule_egress()

    def test__ovn_qos_rule_egress_fip(self):
        self._test__ovn_qos_rule_egress(fip_id='fipid', ip_address='1.2.3.4')

    def test__port_effective_qos_policy_id(self):
        port = {'qos_policy_id': 'qos1'}
        self.assertEqual(('qos1', 'port'),
                         self.qos_driver.port_effective_qos_policy_id(port))

        port = {'qos_network_policy_id': 'qos1'}
        self.assertEqual(('qos1', 'network'),
                         self.qos_driver.port_effective_qos_policy_id(port))

        port = {'qos_policy_id': 'qos_port',
                'qos_network_policy_id': 'qos_network'}
        self.assertEqual(('qos_port', 'port'),
                         self.qos_driver.port_effective_qos_policy_id(port))

        port = {}
        self.assertEqual((None, None),
                         self.qos_driver.port_effective_qos_policy_id(port))

        port = {'qos_policy_id': None, 'qos_network_policy_id': None}
        self.assertEqual((None, None),
                         self.qos_driver.port_effective_qos_policy_id(port))

        port = {'qos_policy_id': 'qos1', 'device_owner': 'neutron:port'}
        self.assertEqual((None, None),
                         self.qos_driver.port_effective_qos_policy_id(port))

    def test_update_port(self):
        port = self.ports[0]
        original_port = self.ports[1]

        # Remove QoS policy
        original_port.qos_policy_id = self.qos_policies[0].id
        self.qos_driver.update_port(mock.ANY, port, original_port)
        self.mock_rules.assert_called_once_with(
            mock.ANY, port.id, port.network_id, None, None)

        # Change from port policy (qos_policy0) to network policy (qos_policy1)
        self.mock_rules.reset_mock()
        port.qos_network_policy_id = self.qos_policies[1].id
        self.qos_driver.update_port(mock.ANY, port, original_port)
        self.mock_rules.assert_called_once_with(
            mock.ANY, port.id, port.network_id, self.qos_policies[1].id, None)

        # No change (qos_policy0)
        self.mock_rules.reset_mock()
        port.qos_policy_id = self.qos_policies[0].id
        original_port.qos_policy_id = self.qos_policies[0].id
        self.qos_driver.update_port(mock.ANY, port, original_port)
        self.mock_rules.assert_not_called()

        # No change (no policy)
        self.mock_rules.reset_mock()
        port.qos_policy_id = None
        port.qos_network_policy_id = None
        original_port.qos_policy_id = None
        original_port.qos_network_policy_id = None
        self.qos_driver.update_port(mock.ANY, port, original_port)
        self.mock_rules.assert_not_called()

        # Reset (no policy)
        self.qos_driver.update_port(mock.ANY, port, original_port, reset=True)
        self.mock_rules.assert_called_once_with(
            mock.ANY, port.id, port.network_id, None, None)

        # Reset (qos_policy0, regardless of being the same a in the previous
        # state)
        self.mock_rules.reset_mock()
        port.qos_policy_id = self.qos_policies[0].id
        original_port.qos_policy_id = self.qos_policies[1].id
        self.qos_driver.update_port(mock.ANY, port, original_port, reset=True)
        self.mock_rules.assert_called_once_with(
            mock.ANY, port.id, port.network_id, self.qos_policies[0].id, None)

        # External port, OVN QoS extension does not apply.
        self.mock_rules.reset_mock()
        port.qos_policy_id = self.qos_policies[0].id
        port_obj.PortBinding(self.ctx, port_id=port.id, host='host',
                             profile={}, vif_type='',
                             vnic_type=portbindings_api.VNIC_DIRECT).create()
        # NOTE(ralonsoh): this OVO retrieval must include, in the port object,
        # the port binding register created.
        port = port_obj.Port.get_object(self.ctx, id=port.id)
        self.qos_driver.update_port(mock.ANY, port, original_port)
        self.mock_rules.assert_not_called()

    def test_delete_port(self):
        self.mock_rules.reset_mock()
        self.qos_driver.delete_port(mock.ANY, self.ports[1])

        # Assert that rules are deleted
        self.mock_rules.assert_called_once_with(
            mock.ANY, self.ports[1].id, self.ports[1].network_id, None, None)

    def test_update_network(self):
        """Test update network.

        net1: [(1) from qos_policy0 to no QoS policy,
               (2) from qos_policy0 to qos_policy1]
        - port10: no QoS port policy
        - port11: qos_policy0
        - port12: qos_policy1
        """
        policies_ports = [
            (None, {self.ports[0].id}),
            (self.qos_policies[1].id, {self.ports[0].id})]

        self.ports[1].qos_policy_id = self.qos_policies[0].id
        self.ports[1].update()
        self.ports[2].qos_policy_id = self.qos_policies[1].id
        self.ports[2].update()
        for qos_policy_id, reference_ports in policies_ports:
            self.networks[0].qos_policy_id = qos_policy_id
            self.networks[0].update()
            original_network = {'qos_policy_id': self.qos_policies[0]}
            reviewed_port_ids, _ = self.qos_driver.update_network(
                mock.ANY, self.networks[0], original_network)
            self.assertEqual(reference_ports, reviewed_port_ids)
            calls = [mock.call(mock.ANY, self.ports[0].id,
                               self.ports[0].network_id, qos_policy_id,
                               None)]
            self.mock_rules.assert_has_calls(calls)
            self.mock_rules.reset_mock()

    def test_update_external_network(self):
        """Test update external network (floating IPs).

        - fip0: qos_policy0
        - fip1: no QoS FIP policy (inherits from external network QoS)
        """
        network_policies = [
            (self.qos_policies[1].id, {self.fips[1].id}),
            (None, {self.fips[1].id})]

        self.fips[0].qos_policy_id = self.qos_policies[0].id
        self.fips[0].update()
        for qos_policy_id, reference_fips in network_policies:
            self.fips_network.qos_policy_id = qos_policy_id
            self.fips_network.update()
            original_network = {'qos_policy_id': self.qos_policies[0]}
            _, reviewed_fips_ids = self.qos_driver.update_network(
                mock.Mock(), self.fips_network, original_network)
            self.assertEqual(reference_fips, reviewed_fips_ids)

    def test_update_network_no_policy_change(self):
        """Test update network if the QoS policy is the same.

        net1: [(1) from qos_policy0 to qos_policy0,
               (2) from no QoS policy to no QoS policy]
        """
        for qos_policy_id in (self.qos_policies[0].id, None):
            self.networks[0].qos_policy_id = qos_policy_id
            self.networks[0].update()
            original_network = {'qos_policy_id': qos_policy_id}
            port_ids, fip_ids = self.qos_driver.update_network(
                mock.ANY, self.networks[0], original_network)
            self.assertEqual(set([]), port_ids)
            self.assertEqual(set([]), fip_ids)
            self.mock_rules.assert_not_called()

    def test_update_network_reset(self):
        """Test update network.

        net1: [(1) from qos_policy1 to qos_policy1,
               (2) from no QoS policy to no QoS policy]
        - port10: no QoS port policy
        - port11: qos_policy0
        - port12: qos_policy1
        """
        policies_ports = [
            (self.qos_policies[1].id, {self.ports[0].id}),
            (None, {self.ports[0].id})]

        self.ports[1].qos_policy_id = self.qos_policies[0].id
        self.ports[1].update()
        self.ports[2].qos_policy_id = self.qos_policies[1].id
        self.ports[2].update()
        for qos_policy_id, reference_ports in policies_ports:
            self.networks[0].qos_policy_id = qos_policy_id
            self.networks[0].update()
            original_network = {'qos_policy_id': self.qos_policies[0]}
            reviewed_port_ids, _ = self.qos_driver.update_network(
                mock.ANY, self.networks[0], original_network, reset=True)
            self.assertEqual(reference_ports, reviewed_port_ids)
            calls = [mock.call(mock.ANY, self.ports[0].id,
                               self.ports[0].network_id, qos_policy_id, None)]
            self.mock_rules.assert_has_calls(calls)
            self.mock_rules.reset_mock()

    def test_update_network_external_ports(self):
        """Test update network with external ports.

        - port10: no QoS port policy
        - port11: no QoS port policy but external
        - port12: qos_policy0
        """
        policies_ports = [(self.qos_policies[0].id, {self.ports[0].id})]
        self.ports[2].qos_policy_id = self.qos_policies[0].id
        self.ports[2].update()
        port_obj.PortBinding(self.ctx, port_id=self.ports[1].id, host='host',
                             profile={}, vif_type='',
                             vnic_type=portbindings_api.VNIC_DIRECT).create()
        with mock.patch.object(self.qos_driver._driver._nb_idl,
                               'get_lswitch_port') as mock_lsp:
            mock_lsp.side_effect = [
                mock.Mock(type=ovn_const.LSP_TYPE_LOCALNET),
                mock.Mock(type=ovn_const.LSP_TYPE_EXTERNAL)]
            for qos_policy_id, reference_ports in policies_ports:
                self.networks[0].qos_policy_id = qos_policy_id
                self.networks[0].update()
                original_network = {'qos_policy_id': self.qos_policies[0]}
                reviewed_port_ids, _ = self.qos_driver.update_network(
                    mock.ANY, self.networks[0], original_network, reset=True)
                self.assertEqual(reference_ports, reviewed_port_ids)
                calls = [mock.call(
                    mock.ANY, self.ports[0].id, self.ports[0].network_id,
                    qos_policy_id, None)]
                self.mock_rules.assert_has_calls(calls)
                self.mock_rules.reset_mock()

    def test_update_policy(self):
        """Test update QoS policy, networks and ports bound are updated.

        QoS policy updated: qos_policy0
        net1: no QoS policy
        - port10: no port QoS policy
        - port11: qos_policy0  --> handled during "update_port" and updated
        - port12: qos_policy1
        net2: qos_policy0
        - port20: no port QoS policy  --> handled during "update_network"
                                          and updated
        - port21: qos_policy0  --> handled during "update_network", not updated
                                   handled during "update_port" and updated
        - port22: qos_policy1  --> handled during "update_network", not updated
        fip1: qos_policy0
        fip2: qos_policy1
        """
        self.ports[1].qos_policy_id = self.qos_policies[0].id
        self.ports[1].update()
        self.ports[2].qos_policy_id = self.qos_policies[1].id
        self.ports[2].update()
        self.ports[4].qos_policy_id = self.qos_policies[0].id
        self.ports[4].update()
        self.ports[5].qos_policy_id = self.qos_policies[1].id
        self.ports[5].update()
        self.networks[1].qos_policy_id = self.qos_policies[0].id
        self.networks[1].update()
        self.fips[0].qos_policy_id = self.qos_policies[0].id
        self.fips[0].update()
        self.fips[1].qos_policy_id = self.qos_policies[1].id
        self.fips[1].update()
        mock_qos_rules = mock.Mock()
        with mock.patch.object(self.qos_driver, '_qos_rules',
                               return_value=mock_qos_rules), \
                mock.patch.object(self.qos_driver, 'update_floatingip') as \
                mock_update_fip:
            self.qos_driver.update_policy(self.ctx, self.qos_policies[0])
        updated_ports = [self.ports[1], self.ports[3], self.ports[4]]
        calls = [mock.call(self.txn, port.id, port.network_id,
                           self.qos_policies[0].id, mock_qos_rules)
                 for port in updated_ports]
        # We can't ensure the call order because we are not enforcing any order
        # when retrieving the port and the network list.
        self.mock_rules.assert_has_calls(calls, any_order=True)
        with db_api.CONTEXT_READER.using(self.ctx):
            fip = self.qos_driver._plugin_l3.get_floatingip(self.ctx,
                                                            self.fips[0].id)
        mock_update_fip.assert_called_once_with(self.txn, fip)

    def test_update_floatingip(self):
        # NOTE(ralonsoh): this rule will always apply:
        # - If the FIP is being deleted, "qos_del_ext_ids" is called;
        #   "qos_add" and "qos_del" won't.
        # - If the FIP is added or updated, "qos_del_ext_ids" won't be called
        #   and "qos_add" or "qos_del" will, depending on the rule directions.
        nb_idl = self.qos_driver._driver._nb_idl
        fip = self.fips[0]
        original_fip = self.fips[1]
        txn = mock.Mock()

        # Update FIP, no QoS policy nor port/router
        self.qos_driver.update_floatingip(txn, fip)
        nb_idl.qos_del_ext_ids.assert_called_once()
        nb_idl.qos_add.assert_not_called()
        nb_idl.qos_del.assert_not_called()
        nb_idl.reset_mock()

        # Attach a port and a router, not QoS policy
        fip.router_id = self.router.id
        fip.fixed_port_id = self.fips_ports[0].id
        fip.update()
        self.qos_driver.update_floatingip(txn, fip)
        nb_idl.qos_del_ext_ids.assert_called_once()
        nb_idl.qos_add.assert_not_called()
        nb_idl.qos_del.assert_not_called()
        nb_idl.reset_mock()

        # Add a QoS policy
        fip.qos_policy_id = self.qos_policies[0].id
        fip.update()
        self.qos_driver.update_floatingip(txn, fip)
        nb_idl.qos_del_ext_ids.assert_not_called()
        # QoS DSCP rule has only egress direction, ingress one is deleted.
        # Check "OVNClientQosExtension.update_floatingip" and how the OVN QoS
        # rules are added (if there is a rule in this direction) or deleted.
        nb_idl.qos_add.assert_called_once()
        nb_idl.qos_del.assert_called_once()
        nb_idl.reset_mock()

        # Remove QoS
        fip.qos_policy_id = None
        fip.update()
        original_fip.qos_policy_id = self.qos_policies[0].id
        original_fip.update()
        self.qos_driver.update_floatingip(txn, fip)
        nb_idl.qos_del_ext_ids.assert_called_once()
        nb_idl.qos_add.assert_not_called()
        nb_idl.qos_del.assert_not_called()
        nb_idl.reset_mock()

        # Add network QoS policy
        fip.qos_network_policy_id = self.qos_policies[0].id
        fip.update()
        self.qos_driver.update_floatingip(txn, fip)
        nb_idl.qos_del_ext_ids.assert_not_called()
        nb_idl.qos_add.assert_called_once()
        nb_idl.qos_del.assert_called_once()
        nb_idl.reset_mock()

        # Add again another QoS policy
        fip.qos_policy_id = self.qos_policies[1].id
        fip.update()
        original_fip.qos_policy_id = None
        original_fip.update()
        self.qos_driver.update_floatingip(txn, fip)
        nb_idl.qos_del_ext_ids.assert_not_called()
        nb_idl.qos_add.assert_called_once()
        nb_idl.qos_del.assert_called_once()
        nb_idl.reset_mock()

        # Detach the port and the router
        fip.router_id = None
        fip.fixed_port_id = None
        fip.update()
        original_fip.router_id = self.router.id
        original_fip.fixed_port_id = self.fips_ports[0].id
        original_fip.qos_policy_id = self.qos_policies[1].id
        original_fip.update()
        self.qos_driver.update_floatingip(txn, fip)
        nb_idl.qos_del_ext_ids.assert_called_once()
        nb_idl.qos_add.assert_not_called()
        nb_idl.qos_del.assert_not_called()
        nb_idl.reset_mock()

        # Force reset (delete any QoS)
        fip_dict = {'floating_network_id': fip.floating_network_id,
                    'id': fip.id}
        self.qos_driver.update_floatingip(txn, fip_dict)
        nb_idl.qos_del_ext_ids.assert_called_once()
        nb_idl.qos_add.assert_not_called()
        nb_idl.qos_del.assert_not_called()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import base as obj_base

from neutron.api.rpc.callbacks import resources
from neutron.objects.qos import policy
from neutron.tests import base


class GetResourceTypeTestCase(base.BaseTestCase):

    def test_get_resource_type_none(self):
        self.assertIsNone(resources.get_resource_type(None))

    def test_get_resource_type_wrong_type(self):
        self.assertIsNone(resources.get_resource_type(object()))

    def test_get_resource_type(self):
        # we could use any other registered NeutronObject type here
        self.assertEqual(policy.QosPolicy.obj_name(),
                         resources.get_resource_type(policy.QosPolicy()))


class IsValidResourceTypeTestCase(base.BaseTestCase):

    def test_known_type(self):
        # it could be any other NeutronObject, assuming it's known to RPC
        # callbacks
        self.assertTrue(resources.is_valid_resource_type(
            policy.QosPolicy.obj_name()))

    def test_unknown_type(self):
        self.assertFalse(
            resources.is_valid_resource_type('unknown-resource-type'))


class GetResourceClsTestCase(base.BaseTestCase):

    def test_known_type(self):
        # it could be any other NeutronObject, assuming it's known to RPC
        # callbacks
        self.assertEqual(policy.QosPolicy,
                         resources.get_resource_cls(resources.QOS_POLICY))

    def test_unknown_type(self):
        self.assertIsNone(resources.get_resource_cls('unknown-resource-type'))


class RegisterResourceClass(base.BaseTestCase):

    def test_register_resource_class(self):
        class DummyOVO(obj_base.VersionedObject):
            pass

        self.assertFalse(
            resources.is_valid_resource_type('DummyOVO'))
        resources.register_resource_class(DummyOVO)
        self.assertTrue(
            resources.is_valid_resource_type('DummyOVO'))

    def test_register_bogus_resource_class(self):
        class DummyOVO(object):
            pass

        self.assertRaises(ValueError,
                          resources.register_resource_class, DummyOVO)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib import exceptions
from neutron_lib.plugins import directory
from oslo_db import exception as db_exc
from oslo_utils import uuidutils

from neutron.ipam import requests as ipam_req
from neutron.ipam import subnet_alloc
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit import testlib_api


class TestSubnetAllocation(testlib_api.SqlTestCase):

    def setUp(self):
        super(TestSubnetAllocation, self).setUp()
        self._tenant_id = 'test-tenant'
        self.setup_coreplugin(test_db_base_plugin_v2.DB_PLUGIN_KLASS)
        self.plugin = directory.get_plugin()
        self.ctx = context.get_admin_context()

    def _create_subnet_pool(self, plugin, ctx, name, prefix_list,
                            min_prefixlen, ip_version,
                            max_prefixlen=constants.ATTR_NOT_SPECIFIED,
                            default_prefixlen=constants.ATTR_NOT_SPECIFIED,
                            default_quota=constants.ATTR_NOT_SPECIFIED,
                            shared=False, is_default=False):
        subnetpool = {'subnetpool': {'name': name,
                                     'tenant_id': self._tenant_id,
                                     'prefixes': prefix_list,
                                     'min_prefixlen': min_prefixlen,
                                     'max_prefixlen': max_prefixlen,
                                     'default_prefixlen': default_prefixlen,
                                     'shared': shared,
                                     'is_default': is_default,
                                     'default_quota': default_quota}}
        return plugin.create_subnetpool(ctx, subnetpool)

    def _get_subnetpool(self, ctx, plugin, id):
        return plugin.get_subnetpool(ctx, id)

    def test_allocate_any_subnet(self):
        prefix_list = ['10.1.0.0/16', '192.168.1.0/24']
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      prefix_list, 21, 4)
        sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
        with db_api.CONTEXT_WRITER.using(self.ctx):
            sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
            req = ipam_req.AnySubnetRequest(self._tenant_id,
                                        uuidutils.generate_uuid(),
                                        constants.IPv4, 21)
            res = sa.allocate_subnet(req)
            detail = res.get_details()
            prefix_set = netaddr.IPSet(iterable=prefix_list)
            allocated_set = netaddr.IPSet(iterable=[detail.subnet_cidr])
            self.assertTrue(allocated_set.issubset(prefix_set))
            self.assertEqual(21, detail.prefixlen)

    def test_allocate_specific_subnet(self):
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['10.1.0.0/16', '192.168.1.0/24'],
                                      21, 4)
        with db_api.CONTEXT_WRITER.using(self.ctx):
            sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
            sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
            req = ipam_req.SpecificSubnetRequest(self._tenant_id,
                                             uuidutils.generate_uuid(),
                                             '10.1.2.0/24')
            res = sa.allocate_subnet(req)
            detail = res.get_details()
            sp = self._get_subnetpool(self.ctx, self.plugin, sp['id'])
            self.assertEqual('10.1.2.0/24', str(detail.subnet_cidr))
            self.assertEqual(24, detail.prefixlen)

    def test_insufficient_prefix_space_for_any_allocation(self):
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['10.1.1.0/24', '192.168.1.0/24'],
                                      21, 4)
        sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
        sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
        req = ipam_req.AnySubnetRequest(self._tenant_id,
                                    uuidutils.generate_uuid(),
                                    constants.IPv4,
                                    21)
        self.assertRaises(exceptions.SubnetAllocationError,
                          sa.allocate_subnet, req)

    def test_insufficient_prefix_space_for_specific_allocation(self):
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['10.1.0.0/24'],
                                      21, 4)
        sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
        sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
        req = ipam_req.SpecificSubnetRequest(self._tenant_id,
                                         uuidutils.generate_uuid(),
                                         '10.1.0.0/21')
        self.assertRaises(exceptions.SubnetAllocationError,
                          sa.allocate_subnet, req)

    def test_allocate_any_subnet_gateway(self):
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['10.1.0.0/16', '192.168.1.0/24'],
                                      21, 4)
        sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
        with db_api.CONTEXT_WRITER.using(self.ctx):
            sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
            req = ipam_req.AnySubnetRequest(self._tenant_id,
                                        uuidutils.generate_uuid(),
                                        constants.IPv4, 21)
            res = sa.allocate_subnet(req)
            detail = res.get_details()
            self.assertEqual(detail.gateway_ip,
                             detail.subnet_cidr.network + 1)

    def test_allocate_specific_subnet_specific_gateway(self):
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['10.1.0.0/16', '192.168.1.0/24'],
                                      21, 4)
        sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
        with db_api.CONTEXT_WRITER.using(self.ctx):
            sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
            req = ipam_req.SpecificSubnetRequest(self._tenant_id,
                                             uuidutils.generate_uuid(),
                                             '10.1.2.0/24',
                                             gateway_ip='10.1.2.254')
            res = sa.allocate_subnet(req)
            detail = res.get_details()
            self.assertEqual(netaddr.IPAddress('10.1.2.254'),
                             detail.gateway_ip)

    def test_allocate_specific_ipv6_subnet_specific_gateway(self):
        # Same scenario as described in bug #1466322
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['2210::/64'],
                                      64, 6)
        sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
        with db_api.CONTEXT_WRITER.using(self.ctx):
            sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
            req = ipam_req.SpecificSubnetRequest(self._tenant_id,
                                                 uuidutils.generate_uuid(),
                                                 '2210::/64',
                                                 '2210::ffff:ffff:ffff:ffff')
            res = sa.allocate_subnet(req)
            detail = res.get_details()
            self.assertEqual(netaddr.IPAddress('2210::ffff:ffff:ffff:ffff'),
                             detail.gateway_ip)

    def test__allocation_value_for_tenant_no_allocations(self):
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['10.1.0.0/16', '192.168.1.0/24'],
                                      21, 4)
        sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
        value = sa._allocations_used_by_tenant(32)
        self.assertEqual(0, value)

    def test_subnetpool_default_quota_exceeded(self):
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['fe80::/48'],
                                      48, 6, default_quota=1)
        sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
        sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
        req = ipam_req.SpecificSubnetRequest(self._tenant_id,
                                         uuidutils.generate_uuid(),
                                         'fe80::/63')
        self.assertRaises(exceptions.SubnetPoolQuotaExceeded,
                          sa.allocate_subnet,
                          req)

    def test_subnetpool_concurrent_allocation_exception(self):
        sp = self._create_subnet_pool(self.plugin, self.ctx, 'test-sp',
                                      ['fe80::/48'],
                                      48, 6, default_quota=1)
        sp = self.plugin._get_subnetpool(self.ctx, sp['id'])
        sa = subnet_alloc.SubnetAllocator(sp, self.ctx)
        req = ipam_req.SpecificSubnetRequest(self._tenant_id,
                                         uuidutils.generate_uuid(),
                                         'fe80::/63')
        with mock.patch("sqlalchemy.orm.query.Query.update", return_value=0):
            self.assertRaises(db_exc.RetryRequest, sa.allocate_subnet, req)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base

DEPRECATED_REASON = (
    "The subnet API now supports system scope and default roles.")

COLLECTION_PATH = '/subnets'
RESOURCE_PATH = '/subnets/{id}'

ACTION_POST = [
    {'method': 'POST', 'path': COLLECTION_PATH},
]
ACTION_PUT = [
    {'method': 'PUT', 'path': RESOURCE_PATH},
]
ACTION_DELETE = [
    {'method': 'DELETE', 'path': RESOURCE_PATH},
]
ACTION_GET = [
    {'method': 'GET', 'path': COLLECTION_PATH},
    {'method': 'GET', 'path': RESOURCE_PATH},
]


rules = [
    policy.DocumentedRuleDefault(
        name='create_subnet',
        check_str=base.policy_or(
            base.PROJECT_MEMBER,
            base.RULE_NET_OWNER),
        scope_types=['project'],
        description='Create a subnet',
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_subnet',
            check_str=base.RULE_ADMIN_OR_NET_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_subnet:segment_id',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description=(
            'Specify ``segment_id`` attribute when creating a subnet'
        ),
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_subnet:segment_id',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_subnet:service_types',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description=(
            'Specify ``service_types`` attribute when creating a subnet'
        ),
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_subnet:service_types',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_subnet',
        check_str=base.policy_or(
            base.PROJECT_READER,
            'rule:shared'),
        scope_types=['project'],
        description='Get a subnet',
        operations=ACTION_GET,
        deprecated_rule=policy.DeprecatedRule(
            name='get_subnet',
            check_str=base.policy_or(
                base.RULE_ADMIN_OR_OWNER,
                'rule:shared'),
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_subnet:segment_id',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Get ``segment_id`` attribute of a subnet',
        operations=ACTION_GET,
        deprecated_rule=policy.DeprecatedRule(
            name='get_subnet:segment_id',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_subnet',
        check_str=base.policy_or(
            base.PROJECT_MEMBER,
            base.RULE_NET_OWNER),
        scope_types=['project'],
        description='Update a subnet',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_subnet',
            check_str=base.RULE_ADMIN_OR_NET_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_subnet:segment_id',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Update ``segment_id`` attribute of a subnet',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_subnet:segment_id',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_subnet:service_types',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Update ``service_types`` attribute of a subnet',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_subnet:service_types',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_subnet',
        check_str=base.policy_or(
            base.PROJECT_MEMBER,
            base.RULE_NET_OWNER),
        scope_types=['project'],
        description='Delete a subnet',
        operations=ACTION_DELETE,
        deprecated_rule=policy.DeprecatedRule(
            name='delete_subnet',
            check_str=base.RULE_ADMIN_OR_NET_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
]


def list_rules():
    return rules



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron_lib.db import model_base
from neutron_lib.db import standard_attr
import sqlalchemy as sa
from sqlalchemy import orm


class Tag(model_base.BASEV2):
    standard_attr_id = sa.Column(
        sa.BigInteger().with_variant(sa.Integer(), 'sqlite'),
        sa.ForeignKey(standard_attr.StandardAttribute.id, ondelete="CASCADE"),
        nullable=False, primary_key=True)
    tag = sa.Column(sa.String(255), nullable=False, primary_key=True)
    standard_attr = orm.relationship(
        'StandardAttribute', load_on_pending=True,
        backref=orm.backref('tags', lazy='subquery', viewonly=True),
        sync_backref=False)
    revises_on_change = ('standard_attr', )



# Copyright 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa


"""add propagate_uplink_status to port

Revision ID: cada2437bf41
Revises: d72db3e25539
Create Date: 2018-11-29 19:25:12.197590

"""

# revision identifiers, used by Alembic.
revision = 'cada2437bf41'
down_revision = 'd72db3e25539'


def upgrade():
    op.create_table('portuplinkstatuspropagation',
                    sa.Column('port_id', sa.String(36),
                              sa.ForeignKey('ports.id',
                                            ondelete="CASCADE"),
                              primary_key=True, index=True),
                    sa.Column('propagate_uplink_status', sa.Boolean(),
                              nullable=False,
                              server_default=sa.sql.false()))



# Copyright (c) 2017 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
from unittest import mock

from neutron_lib import constants
from neutron_lib import context as neutron_context
from oslo_config import cfg
from oslo_log import log as logging
import testscenarios

from neutron.objects.logapi import logging_resource as log_object
from neutron.plugins.ml2.drivers.openvswitch.agent import (
    ovs_agent_extension_api as ovs_ext_api)
from neutron.plugins.ml2.drivers.openvswitch.agent.common import (
    constants as ovs_consts)
from neutron.services.logapi.drivers.openvswitch import (
    ovs_firewall_log as ovs_fw_log)
from neutron.tests.functional.agent import test_firewall

LOG = logging.getLogger(__name__)

load_tests = testscenarios.load_tests_apply_scenarios

FAKE_LOG_ID = 'a2d72369-4246-4f19-bd3c-af51ec8d70cd'
FAKE_PROJECT_ID = 'fake_project'

log_object_dict = {
    'id': FAKE_LOG_ID,
    'resource_type': 'security_group',
    'project_id': FAKE_PROJECT_ID,
    'event': 'ALL'
}
FAKE_LOG_OBJECT = log_object.Log(**log_object_dict)


class LoggingExtensionTestFramework(test_firewall.BaseFirewallTestCase):

    def setUp(self):
        super(LoggingExtensionTestFramework, self).setUp()
        cfg.CONF.set_override('extensions', ['log'], group='agent')
        self.context = neutron_context.get_admin_context_without_session()
        self._set_resource_rpc_mock()
        if self.firewall_name != 'openvswitch':
            self.skipTest("Logging extension doesn't support firewall driver"
                          " %s at that time " % self.firewall_name)
        self.log_driver = self.initialize_ovs_fw_log()

    def initialize_ovs_fw_log(self):
        self.int_br = ovs_ext_api.OVSCookieBridge(
            self.of_helper.br_int_cls(self.tester.bridge.br_name))
        self.tun_br = self.of_helper.br_tun_cls('br-tun')
        agent_api = ovs_ext_api.OVSAgentExtensionAPI(
            self.int_br, self.tun_br,
            {'physnet1': self.of_helper.br_phys_cls('br-physnet1')})
        log_driver = ovs_fw_log.OVSFirewallLoggingDriver(agent_api)
        log_driver.initialize(self.resource_rpc)
        return log_driver

    def _set_resource_rpc_mock(self):
        self.log_info = []

        def _get_sg_info_mock(context, **kwargs):
            return self.log_info

        self.resource_rpc = mock.patch(
            'neutron.services.logapi.rpc.agent.LoggingApiStub').start()
        self.resource_rpc.get_sg_log_info_for_log_resources.side_effect = (
            _get_sg_info_mock)

    def _set_ports_log(self, sg_rules):
        fake_sg_log_info = [
            {
                'id': FAKE_LOG_ID,
                'ports_log': [
                    {'port_id': self.src_port_desc['device'],
                     'security_group_rules': sg_rules}],
                'event': 'ALL',
                'project_id': FAKE_PROJECT_ID
            }]
        self.log_info = fake_sg_log_info


class TestLoggingExtension(LoggingExtensionTestFramework):

    ip_cidr = '192.168.0.1/24'

    def _is_log_flow_set(self, table, actions):
        flows = self.log_driver.int_br.br.dump_flows_for_table(table)
        pattern = re.compile(
            r"^.* table=%s.* actions=%s" % (table, actions)
        )
        for flow in flows.splitlines():
            if pattern.match(flow.strip()):
                return True
        return False

    def _assert_logging_flows_set(self):
        self.assertTrue(self._is_log_flow_set(
            table=ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE,
            actions=r"resubmit\(,%d\),CONTROLLER:65535" % (
                ovs_consts.ACCEPTED_EGRESS_TRAFFIC_NORMAL_TABLE)))
        self.assertTrue(self._is_log_flow_set(
            table=ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
            actions="CONTROLLER:65535"))
        self.assertTrue(self._is_log_flow_set(
            table=ovs_consts.DROPPED_TRAFFIC_TABLE,
            actions="CONTROLLER:65535"))

    def _assert_logging_flows_not_set(self):
        self.assertFalse(self._is_log_flow_set(
            table=ovs_consts.ACCEPTED_EGRESS_TRAFFIC_TABLE,
            actions=r"resubmit\(,%d\),CONTROLLER:65535" % (
                ovs_consts.ACCEPTED_EGRESS_TRAFFIC_NORMAL_TABLE)))
        self.assertFalse(self._is_log_flow_set(
            table=ovs_consts.ACCEPTED_INGRESS_TRAFFIC_TABLE,
            actions="CONTROLLER:65535"))
        self.assertFalse(self._is_log_flow_set(
            table=ovs_consts.DROPPED_TRAFFIC_TABLE,
            actions="CONTROLLER:65535"))

    def test_log_lifecycle(self):
        sg_rules = [{'ethertype': constants.IPv4,
                     'direction': constants.INGRESS_DIRECTION,
                     'protocol': constants.PROTO_NAME_ICMP,
                     'security_group_id': self.FAKE_SECURITY_GROUP_ID},
                    {'ethertype': constants.IPv4,
                     'direction': constants.EGRESS_DIRECTION,
                     'security_group_id': self.FAKE_SECURITY_GROUP_ID},
                    {'ethertype': constants.IPv6,
                     'protocol': constants.PROTO_NAME_TCP,
                     'port_range_min': 22,
                     'port_range_max': 22,
                     'remote_group_id': 2,
                     'direction': constants.EGRESS_DIRECTION,
                     'security_group_id': self.FAKE_SECURITY_GROUP_ID},
                    ]
        self.firewall.update_security_group_rules(
            self.FAKE_SECURITY_GROUP_ID, sg_rules)
        self.firewall.update_port_filter(self.src_port_desc)
        self._set_ports_log(sg_rules)

        # start log
        self.log_driver.start_logging(
            self.context, log_resources=[FAKE_LOG_OBJECT])
        self._assert_logging_flows_set()

        # stop log
        self.log_driver.stop_logging(
            self.context, log_resources=[FAKE_LOG_OBJECT])
        self._assert_logging_flows_not_set()



# Copyright 2016 Cloudbase Solutions.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netifaces

from neutron.agent.windows import ip_lib
from neutron.tests import base


class TestIpWrapper(base.BaseTestCase):

    def test_get_device_by_ip_no_ip(self):
        ret = ip_lib.IPWrapper().get_device_by_ip(None)
        self.assertIsNone(ret)

    @mock.patch.object(ip_lib.IPWrapper, 'get_devices')
    def test_get_device_by_ip(self, mock_get_devices):
        mock_dev1 = mock.MagicMock()
        mock_dev2 = mock.MagicMock()
        mock_dev1.device_has_ip.return_value = False
        mock_dev2.device_has_ip.return_value = True
        mock_get_devices.return_value = [mock_dev1, mock_dev2]
        ret = ip_lib.IPWrapper().get_device_by_ip('fake_ip')

        self.assertEqual(mock_dev2, ret)

    @mock.patch('netifaces.interfaces')
    def test_get_devices(self, mock_interfaces):
        mock_interfaces.return_value = [mock.sentinel.dev1,
                                        mock.sentinel.dev2]

        ret = ip_lib.IPWrapper().get_devices()
        self.assertEqual(mock.sentinel.dev1, ret[0].name)
        self.assertEqual(mock.sentinel.dev2, ret[1].name)

    @mock.patch('netifaces.interfaces')
    def test_get_devices_error(self, mock_interfaces):
        mock_interfaces.side_effect = OSError
        ret = ip_lib.IPWrapper().get_devices()

        self.assertEqual([], ret)


class TestIpDevice(base.BaseTestCase):

    @mock.patch('netifaces.ifaddresses')
    def test_read_ifaddresses(self, mock_netifaces):
        mock_address = {'addr': mock.sentinel.fake_addr}
        mock_netifaces.return_value = {netifaces.AF_INET: [mock_address]}
        ret = ip_lib.IPDevice("fake_dev").read_ifaddresses()
        self.assertTrue(ret)

    @mock.patch('netifaces.ifaddresses')
    def test_read_ifaddresses_no_ip(self, mock_netifaces):
        mock_netifaces.return_value = {}
        ret = ip_lib.IPDevice("fake_dev").read_ifaddresses()
        self.assertFalse(ret)

    @mock.patch('netifaces.ifaddresses')
    def test_read_ifaddresses_ip_error(self, mock_netifaces):
        mock_netifaces.side_effect = OSError
        ret = ip_lib.IPDevice("fake_dev").read_ifaddresses()
        self.assertFalse(ret)

    @mock.patch('netifaces.ifaddresses')
    def test_read_faddresses_not_found(self, mock_netifaces):
        mock_netifaces.side_effect = ValueError
        ret = ip_lib.IPDevice("fake_dev").read_ifaddresses()
        self.assertFalse(ret)

    def test_device_has_ip(self):
        mock_address = {'addr': mock.sentinel.fake_addr}
        ip_device = ip_lib.IPDevice("fake_dev")
        with mock.patch.object(ip_device, "read_ifaddresses", return_value=(
                {netifaces.AF_INET: [mock_address]})):
            ret = ip_device.device_has_ip(mock.sentinel.fake_addr)
            self.assertTrue(ret)

    def test_device_has_ip_false(self):
        ip_device = ip_lib.IPDevice("fake_dev")
        with mock.patch.object(ip_device, "read_ifaddresses", return_value={}):
            ret = ip_device.device_has_ip(mock.sentinel.fake_addr)
            self.assertFalse(ret)

    def test_device_has_ip_error(self):
        ip_device = ip_lib.IPDevice("fake_dev")
        with mock.patch.object(ip_device, "read_ifaddresses",
                               return_value=None):
            ret = ip_device.device_has_ip(mock.sentinel.fake_addr)
            self.assertFalse(ret)


class TestIPLink(base.BaseTestCase):

    def setUp(self):
        super(TestIPLink, self).setUp()
        parent = ip_lib.IPDevice("fake_dev")
        self.ip_link = ip_lib.IPLink(parent)
        self.ip_link._parent.read_ifaddresses = mock.Mock()

    def test_address(self):
        mock_address = {'addr': mock.sentinel.fake_addr}
        self.ip_link._parent.read_ifaddresses.return_value = {
            netifaces.AF_LINK: [mock_address]}
        self.assertEqual([mock_address['addr']], self.ip_link.address)

    def test_address_no_address(self):
        self.ip_link._parent.read_ifaddresses.return_value = {
            netifaces.AF_LINK: []}
        self.assertEqual([], self.ip_link.address)

    def test_address_error(self):
        self.ip_link._parent.read_ifaddresses.return_value = None
        self.assertFalse(self.ip_link.address)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Initial operations for core resources
from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'networks',
        sa.Column('tenant_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=16), nullable=True),
        sa.Column('admin_state_up', sa.Boolean(), nullable=True),
        sa.Column('shared', sa.Boolean(), nullable=True),
        sa.Column('mtu', sa.Integer(), nullable=True),
        sa.Column('vlan_transparent', sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'ports',
        sa.Column('tenant_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('mac_address', sa.String(length=32), nullable=False),
        sa.Column('admin_state_up', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=16), nullable=False),
        sa.Column('device_id', sa.String(length=255), nullable=False),
        sa.Column('device_owner', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id']),
        sa.UniqueConstraint('network_id', 'mac_address',
                            name='uniq_ports0network_id0mac_address'),
        sa.PrimaryKeyConstraint('id'),
        sa.Index(op.f('ix_ports_network_id_device_owner'), 'network_id',
                 'device_owner'),
        sa.Index(op.f('ix_ports_network_id_mac_address'), 'network_id',
                 'mac_address'))

    op.create_table(
        'subnets',
        sa.Column('tenant_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('network_id', sa.String(length=36), nullable=True),
        sa.Column('ip_version', sa.Integer(), nullable=False),
        sa.Column('cidr', sa.String(length=64), nullable=False),
        sa.Column('gateway_ip', sa.String(length=64), nullable=True),
        sa.Column('enable_dhcp', sa.Boolean(), nullable=True),
        sa.Column('shared', sa.Boolean(), nullable=True),
        sa.Column('ipv6_ra_mode',
                  sa.Enum('slaac', 'dhcpv6-stateful', 'dhcpv6-stateless',
                          name='ipv6_ra_modes'),
                  nullable=True),
        sa.Column('ipv6_address_mode',
                  sa.Enum('slaac', 'dhcpv6-stateful', 'dhcpv6-stateless',
                          name='ipv6_address_modes'),
                  nullable=True),
        sa.Column('subnetpool_id', sa.String(length=36), nullable=True,
                  index=True),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'], ),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'dnsnameservers',
        sa.Column('address', sa.String(length=128), nullable=False),
        sa.Column('subnet_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['subnet_id'], ['subnets.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('address', 'subnet_id'))

    op.create_table(
        'ipallocationpools',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('subnet_id', sa.String(length=36), nullable=True),
        sa.Column('first_ip', sa.String(length=64), nullable=False),
        sa.Column('last_ip', sa.String(length=64), nullable=False),
        sa.ForeignKeyConstraint(['subnet_id'], ['subnets.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'subnetroutes',
        sa.Column('destination', sa.String(length=64), nullable=False),
        sa.Column('nexthop', sa.String(length=64), nullable=False),
        sa.Column('subnet_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['subnet_id'], ['subnets.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('destination', 'nexthop', 'subnet_id'))

    op.create_table(
        'ipallocations',
        sa.Column('port_id', sa.String(length=36), nullable=True),
        sa.Column('ip_address', sa.String(length=64), nullable=False),
        sa.Column('subnet_id', sa.String(length=36), nullable=False),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'], ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['subnet_id'], ['subnets.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('ip_address', 'subnet_id', 'network_id'))

    op.create_table(
        'ipavailabilityranges',
        sa.Column('allocation_pool_id', sa.String(length=36), nullable=False),
        sa.Column('first_ip', sa.String(length=64), nullable=False),
        sa.Column('last_ip', sa.String(length=64), nullable=False),
        sa.ForeignKeyConstraint(['allocation_pool_id'],
                                ['ipallocationpools.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('allocation_pool_id', 'first_ip', 'last_ip'),
        sa.UniqueConstraint(
            'first_ip', 'allocation_pool_id',
            name='uniq_ipavailabilityranges0first_ip0allocation_pool_id'),
        sa.UniqueConstraint(
            'last_ip', 'allocation_pool_id',
            name='uniq_ipavailabilityranges0last_ip0allocation_pool_id'))

    op.create_table(
        'networkdhcpagentbindings',
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('dhcp_agent_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['dhcp_agent_id'], ['agents.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id', 'dhcp_agent_id'))



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools

from neutron.common import utils

from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment
from neutron.tests.unit import testlib_api
from neutron_lib import constants
from oslo_utils import uuidutils

load_tests = testlib_api.module_load_tests


class PortShutDownTest(base.BaseFullStackTestCase):
    # This is a test to confirm the port status
    # on shutting down the port administratively.
    # The port status should no longer be ACTIVE
    # and go to DOWN

    use_dhcp = True
    l2_pop = False
    arp_responder = False
    num_hosts = 1

    scenarios = [
        (constants.AGENT_TYPE_LINUXBRIDGE,
         {'l2_agent_type': constants.AGENT_TYPE_LINUXBRIDGE}),
        (constants.AGENT_TYPE_OVS,
         {'l2_agent_type': constants.AGENT_TYPE_OVS})
    ]

    def setUp(self):
        host_descriptions = [
            environment.HostDescription(
                l2_agent_type=self.l2_agent_type,
                dhcp_agent=self.use_dhcp,
            )
            for _ in range(self.num_hosts)]
        env = environment.Environment(
            environment.EnvironmentDescription(
                l2_pop=self.l2_pop,
                arp_responder=self.arp_responder),
            host_descriptions)
        super(PortShutDownTest, self).setUp(env)

    def _create_external_network_and_subnet(self, tenant_id):
        # This test is not exclusive for the external networks.
        # It is only used here to implicitly create a dhcp port
        # on the network creation.
        network = self.safe_client.create_network(
            tenant_id, name='test-public', external=True, network_type='local')
        self.safe_client.create_subnet(tenant_id, network['id'],
                                       '240.0.0.0/8', gateway_ip='240.0.0.2')
        return network

    def _get_network_dhcp_ports(self, network_id):
        return self.client.list_ports(network_id=network_id,
                          device_owner=constants.DEVICE_OWNER_DHCP)['ports']

    def _is_port_active(self, port_id):
        port = self.client.show_port(port_id)['port']
        return port['status'] == constants.PORT_STATUS_ACTIVE

    def _is_port_down(self, port_id):
        port = self.client.show_port(port_id)['port']
        return port['status'] == constants.PORT_STATUS_DOWN

    def test_port_shut_down(self):
        tenant_id = uuidutils.generate_uuid()
        # Create an external network
        network = self._create_external_network_and_subnet(tenant_id)

        # Check if the DHCP port is created
        port_created = functools.partial(self._get_network_dhcp_ports,
                                         network['id'])
        utils.wait_until_true(port_created)

        # Get the DHCP port
        port = self._get_network_dhcp_ports(network['id'])[0]

        # Wait till the changes are reflected to DB
        port_status_active_predicate = functools.partial(
            self._is_port_active, port['id'])
        utils.wait_until_true(port_status_active_predicate)

        # Shut down the port
        self.safe_client.update_port(port['id'], admin_state_up=False)

        port_status_down_predicate = functools.partial(
            self._is_port_down, port['id'])
        utils.wait_until_true(port_status_down_predicate)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import warnings

from neutron_lib.api import converters
from neutron_lib.api import extensions as api_extensions
from neutron_lib.api import faults
from neutron_lib.db import constants as const
from neutron_lib import exceptions
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
import webob

from neutron._i18n import _
from neutron.api import extensions
from neutron.api.v2 import base
from neutron.api.v2 import resource
from neutron import policy
from neutron import quota
from neutron.quota import resource_registry
from neutron import wsgi


DEFAULT_QUOTAS_ACTION = 'default'
RESOURCE_NAME = 'quota'
RESOURCE_COLLECTION = RESOURCE_NAME + "s"
QUOTAS = quota.QUOTAS
DB_QUOTA_DRIVER = cfg.CONF.QUOTAS.quota_driver
EXTENDED_ATTRIBUTES_2_0 = {
    RESOURCE_COLLECTION: {}
}
LOG = logging.getLogger(__name__)


def validate_policy(context, policy_name):
    policy.init()
    policy.enforce(context,
                   policy_name,
                   target={'project_id': context.project_id},
                   plugin=None)


class QuotaSetsController(wsgi.Controller):

    def __init__(self, plugin):
        self._resource_name = RESOURCE_NAME
        self._plugin = plugin
        self._driver = importutils.import_class(cfg.CONF.QUOTAS.quota_driver)()
        self._update_extended_attributes = True

    def _update_attributes(self):
        for quota_resource in resource_registry.get_all_resources().keys():
            attr_dict = EXTENDED_ATTRIBUTES_2_0[RESOURCE_COLLECTION]
            attr_dict[quota_resource] = {
                'allow_post': False,
                'allow_put': True,
                'convert_to': converters.convert_to_int,
                'validate': {'type:range': [-1, const.DB_INTEGER_MAX_VALUE]},
                'is_visible': True}
        self._update_extended_attributes = False

    def _get_quotas(self, request, project_id):
        return self._driver.get_project_quotas(
            request.context,
            resource_registry.get_all_resources(),
            project_id)

    def default(self, request, id):
        context = request.context
        if id != context.project_id:
            validate_policy(context, "get_quota")
        return {self._resource_name: self._driver.get_default_quotas(
                   context=context,
                   resources=resource_registry.get_all_resources(),
                   project_id=id)}

    def create(self, request, body=None):
        msg = _('POST requests are not supported on this resource.')
        raise webob.exc.HTTPNotImplemented(msg)

    def index(self, request):
        context = request.context
        validate_policy(context, "get_quota")
        return {self._resource_name + "s":
                self._driver.get_all_quotas(
                    context, resource_registry.get_all_resources())}

    def tenant(self, request):
        """Retrieve the project info in context."""
        warnings.warn(
            '"tenant" Quota API method is deprecated, use "project" instead')
        return self._project(request, 'tenant')

    def project(self, request):
        """Retrieve the project info in context."""
        return self._project(request, 'project')

    @staticmethod
    def _project(request, key):
        """Retrieve the project info in context."""
        context = request.context
        if not context.project_id:
            raise exceptions.QuotaMissingTenant()
        return {key: {key + '_id': context.project_id}}

    def show(self, request, id):
        if id != request.context.project_id:
            validate_policy(request.context, "get_quota")
        return {self._resource_name: self._get_quotas(request, id)}

    def delete(self, request, id):
        validate_policy(request.context, "delete_quota")
        self._driver.delete_project_quota(request.context, id)

    def update(self, request, id, body=None):
        validate_policy(request.context, "update_quota")
        force = body[self._resource_name].pop('force', None)
        check_limit = body[self._resource_name].pop('check_limit', None)
        # NOTE(ralonsoh): these warning messages will be removed once
        # LP#1953170 is completed and Neutron quota engine accepts "--force" or
        # nothing (by default, Neutron quota engine will check the resource
        # usage before setting the quota limit).
        if force is None and check_limit is None:
            warnings.warn('Neutron quota engine will require "--force" '
                          'parameter to set a quota limit without checking '
                          'the resource usage.')
        elif check_limit:
            warnings.warn('"--check-limit" parameter will not be needed in '
                          'Z+. By default, Neutron quota engine will check '
                          'the resource usage before setting a new quota '
                          'limit. Use "--force" to skip this check.')

        if self._update_extended_attributes:
            self._update_attributes()
        try:
            body = base.Controller.prepare_request_body(
                request.context, body, False, self._resource_name,
                EXTENDED_ATTRIBUTES_2_0[RESOURCE_COLLECTION])
        except Exception as e:
            LOG.warning(
                "An exception happened while processing the request "
                "body. The exception message is [%s].", e)
            raise e

        if check_limit:
            resources = resource_registry.get_all_resources()
            for resource_name, limit in body[self._resource_name].items():
                resource_usage = self._driver.get_resource_usage(
                    request.context, id, resources, resource_name)
                if resource_usage > limit:
                    msg = ('Quota limit %(limit)s for %(resource)s must be '
                           'greater than or equal to already used '
                           '%(resource_usage)s' %
                           {'limit': limit, 'resource': resource_name,
                            'resource_usage': resource_usage})
                    raise webob.exc.HTTPBadRequest(msg)

        for key, value in body[self._resource_name].items():
            self._driver.update_quota_limit(request.context, id, key, value)
        return {self._resource_name: self._get_quotas(request, id)}


class Quotasv2(api_extensions.ExtensionDescriptor):
    """Quotas management support."""

    extensions.register_custom_supported_check(
        RESOURCE_COLLECTION, lambda: True, plugin_agnostic=True)

    @classmethod
    def get_name(cls):
        return "Quota management support"

    @classmethod
    def get_alias(cls):
        return RESOURCE_COLLECTION

    @classmethod
    def get_description(cls):
        description = 'Expose functions for quotas management'
        if cfg.CONF.QUOTAS.quota_driver == DB_QUOTA_DRIVER:
            description += ' per project'
        return description

    @classmethod
    def get_updated(cls):
        return "2012-07-29T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        controller = resource.Resource(
            QuotaSetsController(directory.get_plugin()),
            faults=faults.FAULT_MAP)
        return [extensions.ResourceExtension(
            Quotasv2.get_alias(),
            controller,
            member_actions={DEFAULT_QUOTAS_ACTION: 'GET'},
            collection_actions={'tenant': 'GET',
                                'project': 'GET'})]

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

import os
import signal
import textwrap
from unittest import mock

from neutron_lib import constants as n_consts
from oslo_config import cfg
from oslo_utils import uuidutils
import testtools

from neutron.agent.linux import external_process
from neutron.agent.linux import keepalived
from neutron.conf.agent.l3 import config as l3_config
from neutron.tests import base

# Keepalived user guide:
# http://www.keepalived.org/pdf/UserGuide.pdf

KEEPALIVED_GLOBAL_CONFIG = textwrap.dedent("""\
    global_defs {
        notification_email_from %(email_from)s
        router_id %(router_id)s
    }""") % dict(
        email_from=keepalived.KEEPALIVED_EMAIL_FROM,
        router_id=keepalived.KEEPALIVED_ROUTER_ID)
VRRP_ID = 1
VRRP_INTERVAL = 5


class KeepalivedBaseTestCase(base.BaseTestCase):

    def setUp(self):
        super(KeepalivedBaseTestCase, self).setUp()
        l3_config.register_l3_agent_config_opts(l3_config.OPTS, cfg.CONF)
        self._mock_no_track_supported = mock.patch.object(
            keepalived, '_is_keepalived_use_no_track_supported')


class KeepalivedGetFreeRangeTestCase(KeepalivedBaseTestCase):
    def test_get_free_range(self):
        free_range = keepalived.get_free_range(
            parent_range='169.254.0.0/16',
            excluded_ranges=['169.254.0.0/24',
                             '169.254.1.0/24',
                             '169.254.2.0/24'],
            size=24)
        self.assertEqual('169.254.3.0/24', free_range)

    def test_get_free_range_without_excluded(self):
        free_range = keepalived.get_free_range(
            parent_range='169.254.0.0/16',
            excluded_ranges=[],
            size=20)
        self.assertEqual('169.254.0.0/20', free_range)

    def test_get_free_range_excluded_out_of_parent(self):
        free_range = keepalived.get_free_range(
            parent_range='169.254.0.0/16',
            excluded_ranges=['255.255.255.0/24'],
            size=24)
        self.assertEqual('169.254.0.0/24', free_range)

    def test_get_free_range_not_found(self):
        tiny_parent_range = '192.168.1.0/24'
        huge_size = 8
        with testtools.ExpectedException(ValueError):
            keepalived.get_free_range(
                parent_range=tiny_parent_range,
                excluded_ranges=[],
                size=huge_size)


class KeepalivedConfBaseMixin(object):

    def _get_config(self):
        config = keepalived.KeepalivedConf()

        instance1 = keepalived.KeepalivedInstance('MASTER', 'eth0', 1,
                                                  ['169.254.192.0/18'],
                                                  advert_int=5)
        instance1.set_authentication('AH', 'pass123')
        instance1.track_interfaces.append("eth0")

        vip_address1 = keepalived.KeepalivedVipAddress('192.168.1.0/24',
                                                       'eth1', track=False)

        vip_address2 = keepalived.KeepalivedVipAddress('192.168.2.0/24',
                                                       'eth2', track=False)

        vip_address3 = keepalived.KeepalivedVipAddress('192.168.3.0/24',
                                                       'eth2', track=False)

        vip_address_ex = keepalived.KeepalivedVipAddress('192.168.55.0/24',
                                                         'eth10', track=False)

        instance1.vips.append(vip_address1)
        instance1.vips.append(vip_address2)
        instance1.vips.append(vip_address3)
        instance1.vips.append(vip_address_ex)

        virtual_route = keepalived.KeepalivedVirtualRoute(n_consts.IPv4_ANY,
                                                          "192.168.1.1",
                                                          "eth1")
        instance1.virtual_routes.gateway_routes = [virtual_route]

        instance2 = keepalived.KeepalivedInstance('MASTER', 'eth4', 2,
                                                  ['169.254.192.0/18'],
                                                  mcast_src_ip='224.0.0.1')
        instance2.track_interfaces.append("eth4")

        vip_address1 = keepalived.KeepalivedVipAddress('192.168.3.0/24',
                                                       'eth6', track=False)

        instance2.vips.append(vip_address1)
        instance2.vips.append(vip_address2)
        instance2.vips.append(vip_address_ex)

        config.add_instance(instance1)
        config.add_instance(instance2)

        return config


class KeepalivedConfTestCase(KeepalivedBaseTestCase,
                             KeepalivedConfBaseMixin):

    expected = KEEPALIVED_GLOBAL_CONFIG + textwrap.dedent("""
        vrrp_instance VR_1 {
            state MASTER
            interface eth0
            virtual_router_id 1
            priority 50
            garp_master_delay 60
            advert_int 5
            authentication {
                auth_type AH
                auth_pass pass123
            }
            track_interface {
                eth0
            }
            virtual_ipaddress {
                169.254.0.1/24 dev eth0
            }
            virtual_ipaddress_excluded {
                192.168.1.0/24 dev eth1 no_track
                192.168.2.0/24 dev eth2 no_track
                192.168.3.0/24 dev eth2 no_track
                192.168.55.0/24 dev eth10 no_track
            }
            virtual_routes {
                0.0.0.0/0 via 192.168.1.1 dev eth1 no_track protocol static
            }
        }
        vrrp_instance VR_2 {
            state MASTER
            interface eth4
            virtual_router_id 2
            priority 50
            garp_master_delay 60
            mcast_src_ip 224.0.0.1
            track_interface {
                eth4
            }
            virtual_ipaddress {
                169.254.0.2/24 dev eth4
            }
            virtual_ipaddress_excluded {
                192.168.2.0/24 dev eth2 no_track
                192.168.3.0/24 dev eth6 no_track
                192.168.55.0/24 dev eth10 no_track
            }
        }""")

    def test_config_generation(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=True):
            config = self._get_config()
            self.assertEqual(self.expected, config.get_config_str())

    def test_config_generation_no_track_not_supported(self):
        self._mock_no_track_supported.start().return_value = False
        config = self._get_config()
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=False):
            self.assertEqual(self.expected.replace(' no_track', ''),
                             config.get_config_str())

    def test_config_with_reset(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=True):
            config = self._get_config()
            self.assertEqual(self.expected, config.get_config_str())

        config.reset()
        self.assertEqual(KEEPALIVED_GLOBAL_CONFIG, config.get_config_str())

    def test_get_existing_vip_ip_addresses_returns_list(self):
        config = self._get_config()
        instance = config.get_instance(1)
        current_vips = sorted(instance.get_existing_vip_ip_addresses('eth2'))
        self.assertEqual(['192.168.2.0/24', '192.168.3.0/24'], current_vips)


class KeepalivedConfWithoutNoTrackTestCase(KeepalivedConfTestCase):

    expected = KEEPALIVED_GLOBAL_CONFIG + textwrap.dedent("""
        vrrp_instance VR_1 {
            state MASTER
            interface eth0
            virtual_router_id 1
            priority 50
            garp_master_delay 60
            advert_int 5
            authentication {
                auth_type AH
                auth_pass pass123
            }
            track_interface {
                eth0
            }
            virtual_ipaddress {
                169.254.0.1/24 dev eth0
            }
            virtual_ipaddress_excluded {
                192.168.1.0/24 dev eth1
                192.168.2.0/24 dev eth2
                192.168.3.0/24 dev eth2
                192.168.55.0/24 dev eth10
            }
            virtual_routes {
                0.0.0.0/0 via 192.168.1.1 dev eth1 protocol static
            }
        }
        vrrp_instance VR_2 {
            state MASTER
            interface eth4
            virtual_router_id 2
            priority 50
            garp_master_delay 60
            mcast_src_ip 224.0.0.1
            track_interface {
                eth4
            }
            virtual_ipaddress {
                169.254.0.2/24 dev eth4
            }
            virtual_ipaddress_excluded {
                192.168.2.0/24 dev eth2
                192.168.3.0/24 dev eth6
                192.168.55.0/24 dev eth10
            }
        }""")

    def setUp(self):
        super(KeepalivedConfWithoutNoTrackTestCase, self).setUp()
        cfg.CONF.set_override('keepalived_use_no_track', False)


class KeepalivedStateExceptionTestCase(KeepalivedBaseTestCase):
    def test_state_exception(self):
        invalid_vrrp_state = 'a seal walks'
        self.assertRaises(keepalived.InvalidInstanceStateException,
                          keepalived.KeepalivedInstance,
                          invalid_vrrp_state, 'eth0', 33,
                          ['169.254.192.0/18'])

        invalid_auth_type = 'into a club'
        instance = keepalived.KeepalivedInstance('MASTER', 'eth0', 1,
                                                 ['169.254.192.0/18'])
        self.assertRaises(keepalived.InvalidAuthenticationTypeException,
                          instance.set_authentication,
                          invalid_auth_type, 'some_password')


class KeepalivedInstanceRoutesTestCase(KeepalivedBaseTestCase):
    @classmethod
    def _get_instance_routes(cls):
        routes = keepalived.KeepalivedInstanceRoutes()
        default_gw_eth0 = keepalived.KeepalivedVirtualRoute(
            '0.0.0.0/0', '1.0.0.254', 'eth0')
        default_gw_eth1 = keepalived.KeepalivedVirtualRoute(
            '::/0', 'fe80::3e97:eff:fe26:3bfa/64', 'eth1')
        routes.gateway_routes = [default_gw_eth0, default_gw_eth1]
        extra_routes = [
            keepalived.KeepalivedVirtualRoute('10.0.0.0/8', '1.0.0.1'),
            keepalived.KeepalivedVirtualRoute('20.0.0.0/8', '2.0.0.2')]
        routes.extra_routes = extra_routes
        extra_subnets = [
            keepalived.KeepalivedVirtualRoute(
                '30.0.0.0/8', None, 'eth0', scope='link')]
        routes.extra_subnets = extra_subnets
        return routes

    def test_routes(self):
        routes = self._get_instance_routes()
        self.assertEqual(len(routes.routes), 5)

    def test_remove_routes_on_interface(self):
        routes = self._get_instance_routes()
        routes.remove_routes_on_interface('eth0')
        self.assertEqual(len(routes.routes), 3)
        routes.remove_routes_on_interface('eth1')
        self.assertEqual(len(routes.routes), 2)

    def test_build_config(self):
        expected = """    virtual_routes {
        0.0.0.0/0 via 1.0.0.254 dev eth0 no_track protocol static
        ::/0 via fe80::3e97:eff:fe26:3bfa/64 dev eth1 no_track protocol static
        10.0.0.0/8 via 1.0.0.1 no_track protocol static
        20.0.0.0/8 via 2.0.0.2 no_track protocol static
        30.0.0.0/8 dev eth0 scope link no_track protocol static
    }"""
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=True):
            routes = self._get_instance_routes()
            self.assertEqual(expected, '\n'.join(routes.build_config()))

    def _get_no_track_less_expected_config(self):
        expected = """    virtual_routes {
        0.0.0.0/0 via 1.0.0.254 dev eth0 protocol static
        ::/0 via fe80::3e97:eff:fe26:3bfa/64 dev eth1 protocol static
        10.0.0.0/8 via 1.0.0.1 protocol static
        20.0.0.0/8 via 2.0.0.2 protocol static
        30.0.0.0/8 dev eth0 scope link protocol static
    }"""
        return expected

    def test_build_config_no_track_not_supported(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=False):
            routes = self._get_instance_routes()
            self.assertEqual(self._get_no_track_less_expected_config(),
                             '\n'.join(routes.build_config()))

    def test_build_config_without_no_track_option(self):
        cfg.CONF.set_override('keepalived_use_no_track', False)
        routes = self._get_instance_routes()
        self.assertEqual(self._get_no_track_less_expected_config(),
                         '\n'.join(routes.build_config()))


class KeepalivedInstanceTestCase(KeepalivedBaseTestCase,
                                 KeepalivedConfBaseMixin):
    def test_get_primary_vip(self):
        instance = keepalived.KeepalivedInstance('MASTER', 'ha0', 42,
                                                 ['169.254.192.0/18'])
        self.assertEqual('169.254.0.42/24', instance.get_primary_vip())

    def _test_remove_addresses_by_interface(self, no_track_value):
        config = self._get_config()
        instance = config.get_instance(1)
        instance.remove_vips_vroutes_by_interface('eth2')
        instance.remove_vips_vroutes_by_interface('eth10')

        expected = KEEPALIVED_GLOBAL_CONFIG + textwrap.dedent("""
            vrrp_instance VR_1 {
                state MASTER
                interface eth0
                virtual_router_id 1
                priority 50
                garp_master_delay 60
                advert_int 5
                authentication {
                    auth_type AH
                    auth_pass pass123
                }
                track_interface {
                    eth0
                }
                virtual_ipaddress {
                    169.254.0.1/24 dev eth0
                }
                virtual_ipaddress_excluded {
                    192.168.1.0/24 dev eth1%(no_track)s
                }
                virtual_routes {
                    0.0.0.0/0 via 192.168.1.1 dev eth1%(no_track)s protocol static
                }
            }
            vrrp_instance VR_2 {
                state MASTER
                interface eth4
                virtual_router_id 2
                priority 50
                garp_master_delay 60
                mcast_src_ip 224.0.0.1
                track_interface {
                    eth4
                }
                virtual_ipaddress {
                    169.254.0.2/24 dev eth4
                }
                virtual_ipaddress_excluded {
                    192.168.2.0/24 dev eth2%(no_track)s
                    192.168.3.0/24 dev eth6%(no_track)s
                    192.168.55.0/24 dev eth10%(no_track)s
                }
            }""" % {'no_track': no_track_value})  # noqa: E501 # pylint: disable=line-too-long

        self.assertEqual(expected, config.get_config_str())

    def test_remove_addresses_by_interface(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=True):
            self._test_remove_addresses_by_interface(" no_track")

    def test_remove_address_by_interface_no_track_not_supported(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=False):
            self._test_remove_addresses_by_interface("")

    def test_remove_addresses_by_interface_without_no_track(self):
        cfg.CONF.set_override('keepalived_use_no_track', False)
        self._test_remove_addresses_by_interface("")

    def test_build_config_no_vips(self):
        expected = textwrap.dedent("""\
            vrrp_instance VR_1 {
                state MASTER
                interface eth0
                virtual_router_id 1
                priority 50
                garp_master_delay 60
                virtual_ipaddress {
                    169.254.0.1/24 dev eth0
                }
            }""")
        instance = keepalived.KeepalivedInstance(
            'MASTER', 'eth0', VRRP_ID, ['169.254.192.0/18'])
        self.assertEqual(expected, os.linesep.join(instance.build_config()))

    def test_build_config_no_vips_track_script(self):
        expected = """
vrrp_script ha_health_check_1 {
    script "/etc/ha_confs/qrouter-x/ha_check_script_1.sh"
    interval 5
    fall 2
    rise 2
}

vrrp_instance VR_1 {
    state MASTER
    interface eth0
    virtual_router_id 1
    priority 50
    garp_master_delay 60
    virtual_ipaddress {
        169.254.0.1/24 dev eth0
    }
}"""
        instance = keepalived.KeepalivedInstance(
            'MASTER', 'eth0', VRRP_ID, ['169.254.192.0/18'])
        instance.track_script = keepalived.KeepalivedTrackScript(
            VRRP_INTERVAL, '/etc/ha_confs/qrouter-x', VRRP_ID)
        self.assertEqual(expected, '\n'.join(instance.build_config()))


class KeepalivedVipAddressTestCase(KeepalivedBaseTestCase):
    def test_vip_with_scope(self):
        vip = keepalived.KeepalivedVipAddress('fe80::3e97:eff:fe26:3bfa/64',
                                              'eth1',
                                              'link')
        self.assertEqual('fe80::3e97:eff:fe26:3bfa/64 dev eth1 scope link',
                         vip.build_config())

    def test_add_vip_idempotent(self):
        instance = keepalived.KeepalivedInstance('MASTER', 'eth0', 1,
                                                 ['169.254.192.0/18'])
        instance.add_vip('192.168.222.1/32', 'eth11', None)
        instance.add_vip('192.168.222.1/32', 'eth12', 'link')
        self.assertEqual(1, len(instance.vips))


class KeepalivedVirtualRouteTestCase(KeepalivedBaseTestCase):
    def test_virtual_route_with_dev(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=True):
            route = keepalived.KeepalivedVirtualRoute(
                n_consts.IPv4_ANY, '1.2.3.4', 'eth0')
            self.assertEqual(
                '0.0.0.0/0 via 1.2.3.4 dev eth0 no_track protocol static',
                route.build_config())

    def test_virtual_route_with_dev_no_track_not_supported(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=False):
            route = keepalived.KeepalivedVirtualRoute(
                n_consts.IPv4_ANY, '1.2.3.4', 'eth0')
            self.assertEqual('0.0.0.0/0 via 1.2.3.4 dev eth0 protocol static',
                             route.build_config())

    def test_virtual_route_with_dev_without_no_track(self):
        cfg.CONF.set_override('keepalived_use_no_track', False)
        route = keepalived.KeepalivedVirtualRoute(n_consts.IPv4_ANY, '1.2.3.4',
                                                  'eth0')
        self.assertEqual('0.0.0.0/0 via 1.2.3.4 dev eth0 protocol static',
                         route.build_config())

    def test_virtual_route_without_dev(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=True):
            route = keepalived.KeepalivedVirtualRoute('50.0.0.0/8', '1.2.3.4')
            self.assertEqual('50.0.0.0/8 via 1.2.3.4 no_track protocol static',
                             route.build_config())

    def test_virtual_route_without_dev_no_track_not_supported(self):
        with mock.patch.object(
                keepalived, '_is_keepalived_use_no_track_supported',
                return_value=False):
            route = keepalived.KeepalivedVirtualRoute('50.0.0.0/8', '1.2.3.4')
            self.assertEqual('50.0.0.0/8 via 1.2.3.4 protocol static',
                             route.build_config())

    def test_virtual_route_without_dev_without_no_track(self):
        cfg.CONF.set_override('keepalived_use_no_track', False)
        route = keepalived.KeepalivedVirtualRoute('50.0.0.0/8', '1.2.3.4')
        self.assertEqual('50.0.0.0/8 via 1.2.3.4 protocol static',
                         route.build_config())


class KeepalivedTrackScriptTestCase(KeepalivedBaseTestCase):

    def test_build_config_preamble(self):
        exp_conf = [
            '',
            'vrrp_script ha_health_check_1 {',
            '    script "/etc/ha_confs/qrouter-x/ha_check_script_1.sh"',
            '    interval 5',
            '    fall 2',
            '    rise 2',
            '}',
            '']
        ts = keepalived.KeepalivedTrackScript(
            VRRP_INTERVAL, '/etc/ha_confs/qrouter-x', VRRP_ID)
        self.assertEqual(exp_conf, ts.build_config_preamble())

    def test_get_config_str(self):
        ts = keepalived.KeepalivedTrackScript(
            VRRP_INTERVAL, '/etc/ha_confs/qrouter-x', VRRP_ID)
        ts.routes = [
            keepalived.KeepalivedVirtualRoute('12.0.0.0/24', '10.0.0.1'), ]
        self.assertEqual('''    track_script {
        ha_health_check_1
    }''', ts.get_config_str())

    def test_get_script_str(self):
        ts = keepalived.KeepalivedTrackScript(
            VRRP_INTERVAL, '/etc/ha_confs/qrouter-x', VRRP_ID)
        ts.routes = [
            keepalived.KeepalivedVirtualRoute('12.0.0.0/24', '10.0.0.1'), ]
        ts.vips = [
            keepalived.KeepalivedVipAddress('192.168.0.3/18', 'ha-xxx'), ]

        self.assertEqual("""#!/bin/bash -eu
ip a | grep 192.168.0.3 || exit 0
ping -c 1 -w 1 10.0.0.1 1>/dev/null || exit 1""",
                         ts._get_script_str())

    def test_get_script_str_no_routes(self):
        ts = keepalived.KeepalivedTrackScript(
            VRRP_INTERVAL, '/etc/ha_confs/qrouter-x', VRRP_ID)

        self.assertEqual('#!/bin/bash -eu\n', ts._get_script_str())

    def test_write_check_script(self):
        conf_dir = '/etc/ha_confs/qrouter-x'
        ts = keepalived.KeepalivedTrackScript(VRRP_INTERVAL, conf_dir, VRRP_ID)
        ts.routes = [
            keepalived.KeepalivedVirtualRoute('12.0.0.0/24', '10.0.0.1'),
            keepalived.KeepalivedVirtualRoute('2001:db8::1', '2001:db8::1'), ]
        with mock.patch.object(keepalived, 'file_utils') as patched_utils:
            ts.write_check_script()
            patched_utils.replace_file.assert_called_with(
                os.path.join(conf_dir, 'ha_check_script_1.sh'),
                """#!/bin/bash -eu

ping -c 1 -w 1 10.0.0.1 1>/dev/null || exit 1
ping6 -c 1 -w 1 2001:db8::1 1>/dev/null || exit 1""",
                0o520
            )

    def test_write_check_script_no_routes(self):
        conf_dir = '/etc/ha_confs/qrouter-x'
        ts = keepalived.KeepalivedTrackScript(
            VRRP_INTERVAL, conf_dir, VRRP_ID)
        with mock.patch.object(keepalived, 'file_utils') as patched_utils:
            ts.write_check_script()
            patched_utils.replace_file.assert_not_called()


class KeepalivedManagerTestCase(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.mock_config = mock.Mock()
        self.mock_config.AGENT.check_child_processes_interval = False
        self.process_monitor = external_process.ProcessMonitor(
            self.mock_config, mock.ANY)
        self.uuid = uuidutils.generate_uuid()
        self.process_monitor.register(
            self.uuid, keepalived.KEEPALIVED_SERVICE_NAME, mock.ANY)
        self.keepalived_manager = keepalived.KeepalivedManager(
            self.uuid, self.mock_config, self.process_monitor, mock.ANY)
        self.mock_get_process = mock.patch.object(self.keepalived_manager,
                                                  'get_process')

    def test_destroy(self):
        mock_get_process = self.mock_get_process.start()
        process = mock.Mock()
        mock_get_process.return_value = process
        process.active = False
        self.keepalived_manager.disable()
        process.disable.assert_called_once_with(
            sig=str(int(signal.SIGTERM)))

    def test_destroy_force(self):
        mock_get_process = self.mock_get_process.start()
        with mock.patch.object(keepalived, 'SIGTERM_TIMEOUT', 0):
            process = mock.Mock()
            mock_get_process.return_value = process
            process.active = True
            self.keepalived_manager.disable()
            process.disable.assert_has_calls([
                mock.call(sig=str(int(signal.SIGTERM))),
                mock.call(sig=str(int(signal.SIGKILL)))])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import constants as common_constants
from neutron_lib import exceptions
from oslo_utils import uuidutils

from neutron.agent.l3 import namespaces
from neutron.agent.linux import ip_lib
from neutron.agent.linux import l3_tc_lib
from neutron.tests.functional import base as functional_base

RATE_LIMIT = 1024
BURST_LIMIT = 512
DEV_NAME = "test_device"


class TcLibTestCase(functional_base.BaseSudoTestCase):

    def create_tc_wrapper_with_namespace_and_device(self):
        ns_name = uuidutils.generate_uuid()
        namespace = namespaces.Namespace(
            ns_name, None,
            mock.Mock(), False)
        namespace.create()
        self.addCleanup(namespace.delete)
        ip_wrapper = ip_lib.IPWrapper(namespace=ns_name)
        tc_device = ip_wrapper.add_tuntap(DEV_NAME)
        tc_device.link.set_up()
        return l3_tc_lib.FloatingIPTcCommand(
            DEV_NAME,
            namespace=ns_name)

    def test_clear_all_filters(self):
        ip_addr = "2.2.2.2"
        l3_tc = self.create_tc_wrapper_with_namespace_and_device()
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)
        l3_tc.set_ip_rate_limit(common_constants.EGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)

        l3_tc.clear_all_filters(common_constants.INGRESS_DIRECTION)
        self.assertRaises(exceptions.FilterIDForIPNotFound,
                          l3_tc.get_filter_id_for_ip,
                          common_constants.INGRESS_DIRECTION,
                          ip_addr)

        l3_tc.clear_all_filters(common_constants.EGRESS_DIRECTION)
        self.assertRaises(exceptions.FilterIDForIPNotFound,
                          l3_tc.get_filter_id_for_ip,
                          common_constants.EGRESS_DIRECTION,
                          ip_addr)

    def test_get_filter_id_for_ip(self):
        ip_addr = "3.3.3.3"
        l3_tc = self.create_tc_wrapper_with_namespace_and_device()
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)
        l3_tc.set_ip_rate_limit(common_constants.EGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)

        self.assertIsNotNone(
            l3_tc.get_filter_id_for_ip(common_constants.INGRESS_DIRECTION,
                                       ip_addr))
        self.assertIsNotNone(
            l3_tc.get_filter_id_for_ip(common_constants.EGRESS_DIRECTION,
                                       ip_addr))

        # testing: IP filter does not exist
        self.assertRaises(exceptions.FilterIDForIPNotFound,
                          l3_tc.get_filter_id_for_ip,
                          common_constants.EGRESS_DIRECTION,
                          '33.33.33.33')

    def test_get_existing_filter_ids(self):
        ip_addr = "4.4.4.4"
        l3_tc = self.create_tc_wrapper_with_namespace_and_device()
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)
        l3_tc.set_ip_rate_limit(common_constants.EGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)

        filter_ids = l3_tc.get_existing_filter_ids(
            common_constants.INGRESS_DIRECTION)
        self.assertNotEqual(0, len(filter_ids))
        filter_ids = l3_tc.get_existing_filter_ids(
            common_constants.EGRESS_DIRECTION)
        self.assertNotEqual(0, len(filter_ids))

    def test_delete_filter_ids(self):
        ip_addr1 = "5.5.5.5"
        ip_addr2 = "6.6.6.6"
        l3_tc = self.create_tc_wrapper_with_namespace_and_device()
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION, ip_addr1,
                                RATE_LIMIT, BURST_LIMIT)
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION, ip_addr2,
                                RATE_LIMIT, BURST_LIMIT)

        filter_ids = l3_tc.get_existing_filter_ids(
            common_constants.INGRESS_DIRECTION)
        self.assertEqual(2, len(filter_ids))
        l3_tc.delete_filter_ids(common_constants.INGRESS_DIRECTION,
                                filter_ids)
        filter_ids = l3_tc.get_existing_filter_ids(
            common_constants.INGRESS_DIRECTION)
        self.assertEqual(0, len(filter_ids))

    def test_set_ip_rate_limit(self):
        ip_addr = "7.7.7.7"
        l3_tc = self.create_tc_wrapper_with_namespace_and_device()
        # Set it multiple times
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION, ip_addr,
                                RATE_LIMIT, BURST_LIMIT)
        # Get only one and no exception
        filter_id = l3_tc.get_filter_id_for_ip(
            common_constants.INGRESS_DIRECTION,
            ip_addr)
        self.assertIsNotNone(filter_id)

    def test_clear_ip_rate_limit(self):
        ip_addr = "8.8.8.8"
        l3_tc = self.create_tc_wrapper_with_namespace_and_device()
        l3_tc.set_ip_rate_limit(common_constants.INGRESS_DIRECTION,
                                ip_addr,
                                RATE_LIMIT, BURST_LIMIT)
        filter_id = l3_tc.get_filter_id_for_ip(
            common_constants.INGRESS_DIRECTION,
            ip_addr)
        self.assertIsNotNone(filter_id)
        filter_id = l3_tc.clear_ip_rate_limit(
            common_constants.INGRESS_DIRECTION,
            ip_addr)
        self.assertIsNone(filter_id)

        # testing: IP filter does not exist
        l3_tc.clear_ip_rate_limit(
            common_constants.INGRESS_DIRECTION,
            "88.88.88.88")



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
from neutron_lib import constants as n_const
from neutron_lib.db import constants as db_const
import sqlalchemy as sa


"""port_numa_affinity_policy

Revision ID: 532aa95457e2
Revises: I38991de2b4
Create Date: 2020-07-10 14:59:18.868245

"""

# revision identifiers, used by Alembic.
revision = '532aa95457e2'
down_revision = 'I38991de2b4'


def upgrade():
    op.create_table('portnumaaffinitypolicies',
                    sa.Column('port_id',
                              sa.String(length=db_const.UUID_FIELD_SIZE),
                              sa.ForeignKey('ports.id', ondelete='CASCADE'),
                              primary_key=True),
                    sa.Column('numa_affinity_policy',
                              sa.Enum(*n_const.PORT_NUMA_POLICIES,
                                      name='numa_affinity_policy'))
                    )



# Copyright (c) 2018 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.plugins import directory
from oslo_utils import importutils

from neutron.objects import router as router_obj
from neutron.services.logapi.common import exceptions as log_exc
from neutron.services.logapi.common import validators
from neutron.tests import base


class FakePlugin(object):

    def __init__(self):
        self.validator_mgr = validators.ResourceValidateRequest.get_instance()
        self.supported_logging_types = ['snat']


class TestSnatLogRequestValidations(base.BaseTestCase):
    """Test validation for SNAT log request"""
    def setUp(self):
        self.log_plugin = FakePlugin()
        importutils.import_module('neutron.services.logapi.common.'
                                  'snat_validate')
        super(TestSnatLogRequestValidations, self).setUp()

    def test_validate_request_resource_id_not_specific(self):
        log_data = {'resource_type': 'snat'}

        with mock.patch.object(directory, 'get_plugin',
                               return_value=self.log_plugin):
            with mock.patch.object(router_obj.Router, 'get_object',
                                   return_value=mock.ANY):
                self.assertRaises(
                    log_exc.ResourceIdNotSpecified,
                    self.log_plugin.validator_mgr.validate_request,
                    mock.ANY,
                    log_data)

    def test_validate_request_resource_id_not_exists(self):
        log_data = {'resource_type': 'snat',
                    'resource_id': 'fake_router_id'}

        with mock.patch.object(directory, 'get_plugin',
                               return_value=self.log_plugin):
            with mock.patch.object(router_obj.Router, 'get_object',
                                   return_value=None):
                self.assertRaises(
                    log_exc.ResourceNotFound,
                    self.log_plugin.validator_mgr.validate_request,
                    mock.ANY,
                    log_data)

    def test_validate_request_with_disable_events(self):
        log_data_1 = {'resource_type': 'snat',
                      'resource_id': 'fake_router_id_1',
                      'event': 'ACCEPT'}
        log_data_2 = {'resource_type': 'snat',
                      'resource_id': 'fake_router_id_2',
                      'event': 'DROP'}
        with mock.patch.object(directory, 'get_plugin',
                               return_value=self.log_plugin):
            with mock.patch.object(router_obj.Router, 'get_object',
                                   return_value=mock.ANY):
                self.assertRaises(
                    log_exc.EventsDisabled,
                    self.log_plugin.validator_mgr.validate_request,
                    mock.ANY,
                    log_data_1)

                self.assertRaises(
                    log_exc.EventsDisabled,
                    self.log_plugin.validator_mgr.validate_request,
                    mock.ANY,
                    log_data_2)

    def test_validate_request_with_snat_disable(self):
        log_data = {'resource_type': 'snat',
                    'resource_id': 'fake_router_id'}
        f_router = mock.Mock()
        f_router.enable_snat = False
        with mock.patch.object(directory, 'get_plugin',
                               return_value=self.log_plugin):
            with mock.patch.object(router_obj.Router, 'get_object',
                                   return_value=f_router):
                self.assertRaises(
                    log_exc.RouterNotEnabledSnat,
                    self.log_plugin.validator_mgr.validate_request,
                    mock.ANY,
                    log_data)

    def test_validate_request_with_not_set_gw_port(self):
        log_data = {'resource_type': 'snat',
                    'resource_id': 'fake_router_id'}
        f_router = mock.Mock()
        f_router.enable_snat = True
        f_router.gw_port_id = None
        with mock.patch.object(directory, 'get_plugin',
                               return_value=self.log_plugin):
            with mock.patch.object(router_obj.Router, 'get_object',
                                   return_value=f_router):
                self.assertRaises(
                    log_exc.RouterGatewayNotSet,
                    self.log_plugin.validator_mgr.validate_request,
                    mock.ANY,
                    log_data)



# Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from neutron_lib.callbacks import events
from neutron_lib.callbacks import resources
from neutron_lib import fixture

from neutron.api.rpc.callbacks import resource_manager
from neutron.services.trunk.rpc import backend
from neutron.tests import base
from neutron.tests import tools


class ServerSideRpcBackendTest(base.BaseTestCase):
    # TODO(fitoduarte): add more test to improve coverage of module
    def setUp(self):
        super(ServerSideRpcBackendTest, self).setUp()
        self._mgr = mock.Mock()
        self.useFixture(fixture.CallbackRegistryFixture(
            callback_manager=self._mgr))
        self.register_mock = mock.patch.object(
            resource_manager.ResourceCallbacksManager, "register").start()

    def test___init__(self,):
        test_obj = backend.ServerSideRpcBackend()

        calls = [mock.call(
                    *tools.get_subscribe_args(
                        test_obj.process_event,
                        resources.TRUNK,
                        events.AFTER_CREATE)),
                 mock.call(
                    *tools.get_subscribe_args(
                        test_obj.process_event,
                        resources.TRUNK,
                        events.AFTER_DELETE)),
                 mock.call(
                    *tools.get_subscribe_args(
                        test_obj.process_event,
                        resources.SUBPORTS,
                        events.AFTER_CREATE)),
                 mock.call(
                    *tools.get_subscribe_args(
                        test_obj.process_event,
                        resources.SUBPORTS,
                        events.AFTER_DELETE))
                 ]
        self._mgr.subscribe.assert_has_calls(calls, any_order=True)

    def test_process_event(self):
        test_obj = backend.ServerSideRpcBackend()
        test_obj._stub = mock_stub = mock.Mock()
        trunk_plugin = mock.Mock()

        test_obj.process_event(
            resources.TRUNK, events.AFTER_CREATE, trunk_plugin,
            events.DBEventPayload("context",
                                  resource_id="id",
                                  states=("current_trunk",)))

        test_obj.process_event(
            resources.TRUNK, events.AFTER_DELETE, trunk_plugin,
            events.DBEventPayload("context",
                                  resource_id="id",
                                  states=("original_trunk",)))

        calls = [mock.call.trunk_created("context",
                                         "current_trunk"),
                 mock.call.trunk_deleted("context",
                                         "original_trunk")]
        mock_stub.assert_has_calls(calls, any_order=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as context_lib
from neutron_lib.objects import common_types
from neutron_lib.utils import net as net_utils
from oslo_utils import versionutils
from oslo_versionedobjects import fields as obj_fields
from sqlalchemy import or_

from neutron.db.models import securitygroup as sg_models
from neutron.db import rbac_db_models
from neutron.objects import base
from neutron.objects import ports
from neutron.objects import rbac
from neutron.objects import rbac_db


@base.NeutronObjectRegistry.register
class SecurityGroupRBAC(rbac.RBACBaseObject):
    # Version 1.0: Initial version
    # Version 1.1: Changed 'target_tenant' to 'target_project'
    VERSION = '1.1'

    db_model = rbac_db_models.SecurityGroupRBAC


@base.NeutronObjectRegistry.register
class SecurityGroup(rbac_db.NeutronRbacObject):
    # Version 1.0: Initial version
    # Version 1.1: Add RBAC support
    # Version 1.2: Added stateful support
    # Version 1.3: Added support for remote_address_group_id in rules
    # Version 1.4: Added support for normalized_cidr in rules
    # Version 1.5: Make the shared field nullable
    VERSION = '1.5'

    # required by RbacNeutronMetaclass
    rbac_db_cls = SecurityGroupRBAC
    db_model = sg_models.SecurityGroup

    fields = {
        'id': common_types.UUIDField(),
        'name': obj_fields.StringField(nullable=True),
        'project_id': obj_fields.StringField(nullable=True),
        'shared': obj_fields.BooleanField(nullable=True),
        'stateful': obj_fields.BooleanField(default=True),
        'is_default': obj_fields.BooleanField(default=False),
        'rules': obj_fields.ListOfObjectsField(
            'SecurityGroupRule', nullable=True
        ),
        # NOTE(ihrachys): we don't include source_rules that is present in the
        # model until we realize it's actually needed
    }

    fields_no_update = ['project_id', 'is_default']

    synthetic_fields = ['is_default', 'rules']

    extra_filter_names = {'is_default'}

    lazy_fields = set(['rules'])

    def create(self):
        # save is_default before super() resets it to False
        is_default = self.is_default
        with self.db_context_writer(self.obj_context):
            super(SecurityGroup, self).create()
            if is_default:
                default_group = DefaultSecurityGroup(
                    self.obj_context,
                    project_id=self.project_id,
                    security_group_id=self.id)
                default_group.create()
                self.is_default = True
                self.obj_reset_changes(['is_default'])

    def from_db_object(self, db_obj):
        super(SecurityGroup, self).from_db_object(db_obj)
        if self._load_synthetic_fields:
            setattr(self, 'is_default',
                    bool(db_obj.get('default_security_group')))
            self.obj_reset_changes(['is_default'])

    @classmethod
    def get_sg_by_id(cls, context, sg_id):
        return super(SecurityGroup, cls).get_object(context, id=sg_id)

    def obj_make_compatible(self, primitive, target_version):
        _target_version = versionutils.convert_version_to_tuple(target_version)

        def filter_remote_address_group_id_from_rules(rules):
            sg_rule = SecurityGroupRule()
            for rule in rules:
                sg_rule.obj_make_compatible(
                    rule['versioned_object.data'], '1.0')
                rule['versioned_object.version'] = '1.0'

        def filter_normalized_cidr_from_rules(rules):
            sg_rule = SecurityGroupRule()
            for rule in rules:
                rule_version = versionutils.convert_version_to_tuple(
                    rule['versioned_object.version'])
                if rule_version > (1, 1):
                    sg_rule.obj_make_compatible(
                        rule['versioned_object.data'], '1.1')
                    rule['versioned_object.version'] = '1.1'

        if _target_version < (1, 1):
            primitive.pop('shared')
        if _target_version < (1, 2):
            primitive.pop('stateful')
        if _target_version < (1, 3):
            if 'rules' in primitive:
                filter_remote_address_group_id_from_rules(primitive['rules'])
        if _target_version < (1, 4):
            if 'rules' in primitive:
                filter_normalized_cidr_from_rules(primitive['rules'])

    @classmethod
    def get_bound_project_ids(cls, context, obj_id):
        port_objs = ports.Port.get_objects(context,
                                           security_group_ids=[obj_id])
        return {port.project_id for port in port_objs}


@base.NeutronObjectRegistry.register
class DefaultSecurityGroup(base.NeutronDbObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    db_model = sg_models.DefaultSecurityGroup

    fields = {
        'project_id': obj_fields.StringField(),
        'security_group_id': common_types.UUIDField(),
    }

    fields_no_update = ['security_group_id']

    primary_keys = ['project_id']


@base.NeutronObjectRegistry.register
class SecurityGroupRule(base.NeutronDbObject):
    # Version 1.0: Initial version
    # Version 1.1: Add remote address group support
    # Version 1.2: Added normalized cidr column
    VERSION = '1.2'

    db_model = sg_models.SecurityGroupRule

    fields = {
        'id': common_types.UUIDField(),
        'project_id': obj_fields.StringField(nullable=True),
        'security_group_id': common_types.UUIDField(),
        'remote_group_id': common_types.UUIDField(nullable=True),
        'direction': common_types.FlowDirectionEnumField(nullable=True),
        'ethertype': common_types.EtherTypeEnumField(nullable=True),
        'protocol': common_types.IpProtocolEnumField(nullable=True),
        'port_range_min': common_types.PortRangeWith0Field(nullable=True),
        'port_range_max': common_types.PortRangeWith0Field(nullable=True),
        'remote_ip_prefix': common_types.IPNetworkField(nullable=True),
        'remote_address_group_id': common_types.UUIDField(nullable=True),
        'normalized_cidr': common_types.IPNetworkField(nullable=True),
    }

    synthetic_fields = ['normalized_cidr']

    foreign_keys = {'SecurityGroup': {'security_group_id': 'id'}}

    fields_no_update = ['project_id', 'security_group_id', 'remote_group_id',
                        'remote_address_group_id']

    def obj_make_compatible(self, primitive, target_version):
        _target_version = versionutils.convert_version_to_tuple(target_version)
        if _target_version < (1, 1):
            primitive.pop('remote_address_group_id', None)
        if _target_version < (1, 2):
            primitive.pop('normalized_cidr', None)

    # TODO(sayalilunkad): get rid of it once we switch the db model to using
    # custom types.
    @classmethod
    def modify_fields_to_db(cls, fields):
        result = super(SecurityGroupRule, cls).modify_fields_to_db(fields)
        remote_ip_prefix = result.get('remote_ip_prefix')
        if remote_ip_prefix:
            result['remote_ip_prefix'] = cls.filter_to_str(remote_ip_prefix)
        return result

    def _load_normalized_cidr(self, db_obj=None):
        db_obj = db_obj or SecurityGroupRule.get_object(self.obj_context,
                                                        id=self.id)
        if not db_obj:
            return

        cidr = None
        if db_obj.remote_ip_prefix:
            cidr = net_utils.AuthenticIPNetwork(db_obj.remote_ip_prefix).cidr

        setattr(self, 'normalized_cidr', cidr)
        self.obj_reset_changes(['normalized_cidr'])

    def from_db_object(self, db_obj):
        super(SecurityGroupRule, self).from_db_object(db_obj)
        self._load_normalized_cidr(db_obj)

    def obj_load_attr(self, attrname):
        if attrname == 'normalized_cidr':
            return self._load_normalized_cidr()
        super(SecurityGroupRule, self).obj_load_attr(attrname)

    # TODO(sayalilunkad): get rid of it once we switch the db model to using
    # custom types.
    @classmethod
    def modify_fields_from_db(cls, db_obj):
        fields = super(SecurityGroupRule, cls).modify_fields_from_db(db_obj)
        if 'remote_ip_prefix' in fields:
            fields['remote_ip_prefix'] = (
                net_utils.AuthenticIPNetwork(fields['remote_ip_prefix']))
        return fields

    @classmethod
    def get_security_group_rule_ids(cls, project_id):
        """Retrieve all SG rules related to this project_id

        This method returns the SG rule IDs that meet these conditions:
        - The rule belongs to this project_id
        - The rule belongs to a security group that belongs to the project_id
        """
        context = context_lib.get_admin_context()
        query = context.session.query(cls.db_model.id)
        query = query.join(
            SecurityGroup.db_model,
            cls.db_model.security_group_id == SecurityGroup.db_model.id)
        clauses = or_(SecurityGroup.db_model.project_id == project_id,
                      cls.db_model.project_id == project_id)
        rule_ids = query.filter(clauses).all()
        return [rule_id[0] for rule_id in rule_ids]



# Copyright 2015 Intel Corporation.
# Copyright 2015 Isaku Yamahata <isaku.yamahata at intel com>
#                               <isaku.yamahata at gmail com>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import netutils

from neutron.agent.linux import bridge_lib
from neutron.privileged.agent.linux import ip_lib as priv_lib
from neutron.tests import base


class BridgeLibTest(base.BaseTestCase):
    """A test suite to exercise the bridge libraries """
    _NAMESPACE = 'test-namespace'
    _BR_NAME = 'test-br'
    _IF_NAME = 'test-if'

    def setUp(self):
        super(BridgeLibTest, self).setUp()
        mock.patch.object(netutils, 'is_ipv6_enabled',
                          return_value=True).start()
        ip_wrapper = mock.patch('neutron.agent.linux.ip_lib.IPWrapper').start()
        self.execute = ip_wrapper.return_value.netns.execute
        self.create_p = mock.patch.object(priv_lib, 'create_interface')
        self.create = self.create_p.start()
        self.delete_p = mock.patch.object(priv_lib, 'delete_interface')
        self.delete = self.delete_p.start()

    def _verify_bridge_mock(self, cmd):
        self.execute.assert_called_once_with(cmd, run_as_root=True)
        self.execute.reset_mock()

    def _verify_bridge_sysctl_mock(self, cmd):
        self.execute.assert_called_once_with(cmd, run_as_root=True,
                                             log_fail_as_error=True)
        self.execute.reset_mock()

    def test_is_bridged_interface(self):
        exists = lambda path: path == "/sys/class/net/tapOK/brport"
        with mock.patch('os.path.exists', side_effect=exists):
            self.assertTrue(bridge_lib.is_bridged_interface("tapOK"))
            self.assertFalse(bridge_lib.is_bridged_interface("tapKO"))

    def test_get_interface_bridge(self):
        with mock.patch('os.readlink', side_effect=["prefix/br0", OSError()]):
            br = bridge_lib.BridgeDevice.get_interface_bridge('tap0')
            self.assertIsInstance(br, bridge_lib.BridgeDevice)
            self.assertEqual("br0", br.name)

            br = bridge_lib.BridgeDevice.get_interface_bridge('tap0')
            self.assertIsNone(br)

    def test_addbr_exists(self):
        self.create.side_effect = priv_lib.InterfaceAlreadyExists
        bridge_lib.BridgeDevice.addbr(self._BR_NAME)
        bridge_lib.BridgeDevice.addbr(self._BR_NAME)

    def test_owns_interface(self):
        br = bridge_lib.BridgeDevice('br-int')
        exists = lambda path: path == "/sys/class/net/br-int/brif/abc"
        with mock.patch('os.path.exists', side_effect=exists):
            self.assertTrue(br.owns_interface("abc"))
            self.assertFalse(br.owns_interface("def"))

    def test_get_interfaces(self):
        br = bridge_lib.BridgeDevice('br-int')
        interfaces = ["tap1", "tap2"]
        with mock.patch('os.listdir', side_effect=[interfaces, OSError()]):
            self.assertEqual(interfaces, br.get_interfaces())
            self.assertEqual([], br.get_interfaces())



# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import weakref

import fixtures
from neutron_lib.plugins import constants as lib_const
from neutron_lib.plugins import directory
from oslo_config import cfg

from neutron import manager
from neutron.plugins.common import constants
from neutron.tests import base
from neutron.tests.unit import dummy_plugin
from neutron.tests.unit import testlib_api


DB_PLUGIN_KLASS = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'


class MultiServiceCorePlugin(object):
    supported_extension_aliases = ['fwaas', dummy_plugin.Dummy.get_alias()]


class CorePluginWithAgentNotifiers(object):
    supported_extension_aliases = []
    agent_notifiers = {'l3': 'l3_agent_notifier',
                       'dhcp': 'dhcp_agent_notifier'}


class NeutronManagerTestCase(base.BaseTestCase):

    def setUp(self):
        ext_mapping = constants.EXT_TO_SERVICE_MAPPING
        if dummy_plugin.Dummy.get_alias() not in ext_mapping:
            ext_mapping[dummy_plugin.Dummy.get_alias()] = (
                dummy_plugin.DUMMY_SERVICE_TYPE)
        super(NeutronManagerTestCase, self).setUp()
        self.config_parse()
        self.setup_coreplugin(load_plugins=False)
        self.useFixture(
            fixtures.MonkeyPatch('neutron.manager.NeutronManager._instance'))

    def tearDown(self):
        ext_mapping = constants.EXT_TO_SERVICE_MAPPING
        if dummy_plugin.Dummy.get_alias() in ext_mapping:
            del ext_mapping[dummy_plugin.Dummy.get_alias()]
        super(NeutronManagerTestCase, self).tearDown()

    def test_service_plugin_is_loaded(self):
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        manager.init()
        plugin = directory.get_plugin(dummy_plugin.DUMMY_SERVICE_TYPE)

        self.assertIsInstance(
            plugin, dummy_plugin.DummyServicePlugin,
            "loaded plugin should be of type neutronDummyPlugin")

    def test_service_plugin_by_name_is_loaded(self):
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        cfg.CONF.set_override("service_plugins",
                              [dummy_plugin.Dummy.get_alias()])
        manager.init()
        plugin = directory.get_plugin(dummy_plugin.DUMMY_SERVICE_TYPE)

        self.assertIsInstance(
            plugin, dummy_plugin.DummyServicePlugin,
            "loaded plugin should be of type neutronDummyPlugin")

    def test_multiple_plugins_specified_for_service_type(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin",
                               "neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        manager.NeutronManager.get_instance()
        plugins = directory.get_plugins()
        # CORE, DUMMY
        self.assertEqual(2, len(plugins))

    def test_multiple_plugins_by_name_specified_for_service_type(self):
        cfg.CONF.set_override("service_plugins",
                              [dummy_plugin.Dummy.get_alias(),
                               dummy_plugin.Dummy.get_alias()])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        manager.NeutronManager.get_instance()
        plugins = directory.get_plugins()
        # CORE, DUMMY
        self.assertEqual(2, len(plugins))

    def test_multiple_plugins_mixed_specified_for_service_type(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin",
                               dummy_plugin.Dummy.get_alias()])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        manager.NeutronManager.get_instance()
        plugins = directory.get_plugins()
        # CORE, DUMMY
        self.assertEqual(2, len(plugins))

    def test_service_plugin_conflicts_with_core_plugin(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        cfg.CONF.set_override("core_plugin",
                              "neutron.tests.unit.test_manager."
                              "MultiServiceCorePlugin")
        manager.NeutronManager.get_instance()
        plugins = directory.get_plugins()
        # CORE, FIREWALL, DUMMY
        self.assertEqual(3, len(plugins))

    def test_load_plugins_with_requirements(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyWithRequireServicePlugin"])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        with mock.patch(
            "neutron.tests.unit.dummy_plugin.DummyServicePlugin.__init__",
            return_value=None
        ) as dummy_init_mock, mock.patch(
            "neutron.tests.unit.dummy_plugin."
            "DummyWithRequireServicePlugin.__init__",
            return_value=None
        ) as dummy_with_require_init_mock:
            manager.NeutronManager.get_instance()
            plugins = directory.get_plugins()
        # DUMMY will also be initialized since DUMMY_REQIURE needs it.
        # CORE, DUMMY, DUMMY_REQIURE
        self.assertEqual(3, len(plugins))
        # ensure that DUMMY and DUMMY_REQIURE was instantiate only once:
        self.assertEqual(1, dummy_init_mock.call_count)
        self.assertEqual(1, dummy_with_require_init_mock.call_count)

    def test_load_plugins_with_requirements_with_parent(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin",
                               "neutron.tests.unit.dummy_plugin."
                               "DummyWithRequireServicePlugin"])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        with mock.patch(
            "neutron.tests.unit.dummy_plugin.DummyServicePlugin.__init__",
            return_value=None
        ) as dummy_init_mock, mock.patch(
            "neutron.tests.unit.dummy_plugin."
            "DummyWithRequireServicePlugin.__init__",
            return_value=None
        ) as dummy_with_require_init_mock:
            manager.NeutronManager.get_instance()
            plugins = directory.get_plugins()
        # CORE, DUMMY, DUMMY_REQIURE
        self.assertEqual(3, len(plugins))
        # ensure that DUMMY and DUMMY_REQIURE was instantiate only once:
        self.assertEqual(1, dummy_init_mock.call_count)
        self.assertEqual(1, dummy_with_require_init_mock.call_count)

    def test_load_plugins_with_requirements_child_first(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyWithRequireServicePlugin",
                               "neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        with mock.patch(
            "neutron.tests.unit.dummy_plugin.DummyServicePlugin.__init__",
            return_value=None
        ) as dummy_init_mock, mock.patch(
            "neutron.tests.unit.dummy_plugin."
            "DummyWithRequireServicePlugin.__init__",
            return_value=None
        ) as dummy_with_require_init_mock:
            manager.NeutronManager.get_instance()
            plugins = directory.get_plugins()
        # CORE, DUMMY, DUMMY_REQIURE
        self.assertEqual(3, len(plugins))
        # ensure that DUMMY and DUMMY_REQIURE was instantiate only once:
        self.assertEqual(1, dummy_init_mock.call_count)
        self.assertEqual(1, dummy_with_require_init_mock.call_count)

    def test_core_plugin_supports_services(self):
        cfg.CONF.set_override("core_plugin",
                              "neutron.tests.unit.test_manager."
                              "MultiServiceCorePlugin")
        manager.init()
        svc_plugins = directory.get_plugins()
        self.assertEqual(3, len(svc_plugins))
        self.assertIn(lib_const.CORE, svc_plugins.keys())
        self.assertIn(dummy_plugin.DUMMY_SERVICE_TYPE, svc_plugins.keys())

    def test_load_default_service_plugins(self):
        self.patched_default_svc_plugins.return_value = {
            'neutron.tests.unit.dummy_plugin.DummyServicePlugin':
                dummy_plugin.DUMMY_SERVICE_TYPE
        }
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        manager.init()
        svc_plugins = directory.get_plugins()
        self.assertIn(dummy_plugin.DUMMY_SERVICE_TYPE, svc_plugins)

    def test_dhcp_agents_per_network_min(self):
        # Ensures dhcp_agents_per_network must be at least 1.
        self.assertRaises(ValueError, cfg.CONF.set_override,
                          'dhcp_agents_per_network', 0)

    def test_pre_plugin_validation(self):
        self.assertIsNotNone(manager.validate_pre_plugin_load())
        cfg.CONF.set_override('core_plugin', 'dummy.plugin')
        self.assertIsNone(manager.validate_pre_plugin_load())

    def test_manager_gathers_agent_notifiers_from_service_plugins(self):
        cfg.CONF.set_override("service_plugins",
                              ["neutron.tests.unit.dummy_plugin."
                               "DummyServicePlugin"])
        cfg.CONF.set_override("core_plugin",
                              "neutron.tests.unit.test_manager."
                              "CorePluginWithAgentNotifiers")
        expected = {'l3': 'l3_agent_notifier',
                    'dhcp': 'dhcp_agent_notifier',
                    dummy_plugin.Dummy.get_alias(): 'dummy_agent_notifier'}
        manager.init()
        core_plugin = directory.get_plugin()
        self.assertEqual(expected, core_plugin.agent_notifiers)

    def test_load_class_for_provider(self):
        manager.NeutronManager.load_class_for_provider(
            'neutron.core_plugins', 'ml2')

    def test_load_class_for_provider_wrong_plugin(self):
        with testlib_api.ExpectedException(ImportError):
            manager.NeutronManager.load_class_for_provider(
                    'neutron.core_plugins', 'ml2XXXXXX')

    def test_get_service_plugin_by_path_prefix_3(self):
        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        nm = manager.NeutronManager.get_instance()

        class pclass(object):
            def __init__(self, path_prefix):
                self.path_prefix = path_prefix

        x_plugin, y_plugin = pclass('xpa'), pclass('ypa')
        directory.add_plugin('x', x_plugin)
        directory.add_plugin('y', y_plugin)
        self.assertEqual(weakref.proxy(x_plugin),
                         nm.get_service_plugin_by_path_prefix('xpa'))
        self.assertEqual(weakref.proxy(y_plugin),
                         nm.get_service_plugin_by_path_prefix('ypa'))
        self.assertIsNone(nm.get_service_plugin_by_path_prefix('abc'))



# Copyright (c) 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as lib_constants

from neutron.agent.l3 import router_info as router
from neutron.agent.linux import ip_lib


class LegacyRouter(router.RouterInfo):
    def add_floating_ip(self, fip, interface_name, device):
        if not self._add_fip_addr_to_device(fip, device):
            return lib_constants.FLOATINGIP_STATUS_ERROR

        # As GARP is processed in a distinct thread the call below
        # won't raise an exception to be handled.
        ip_lib.send_ip_addr_adv_notif(self.ns_name,
                                      interface_name,
                                      fip['floating_ip_address'])
        return lib_constants.FLOATINGIP_STATUS_ACTIVE
